"""Helpers for USD workflows which do not know anything about the pipeline."""
import enum
import typing
import inspect
import logging
import functools
import contextlib

import numpy as np

from itertools import chain
from collections import abc

from pxr import Usd, UsdGeom, Sdf, Plug, Ar, Tf
from printree import TreePrinter

logger = logging.getLogger(__name__)


@functools.cache
def _attr_value_type_names():
    values = inspect.getmembers(Sdf.ValueTypeNames, lambda v: isinstance(v, Sdf.ValueTypeName) and not v.isArray)
    return frozenset(chain.from_iterable(obj.aliasesAsStrings for name, obj in values))


@functools.cache
def _metadata_keys():
    # https://github.com/PixarAnimationStudios/USD/blob/7a5f8c4311fed3ef2271d5e4b51025fb0f513730/pxr/usd/sdf/textFileFormat.yy#L1400-L1409
    keys = {"doc", "subLayers"}
    keys.update(chain.from_iterable(p.metadata.get('SdfMetadata', {}) for p in Plug.Registry().GetAllPlugins()))

    # TODO: investigate if there's another way of doing this, like via the registry above
    stage = Usd.Stage.CreateInMemory()
    UsdGeom.Scope.Define(stage, "/a").MakeInvisible()

    layer = stage.GetRootLayer()
    layer.Traverse(layer.pseudoRoot.path, lambda path: keys.update(layer.GetObjectAtPath(path).GetMetaDataInfoKeys()))
    return frozenset(keys)


def _pruned_prims(prim_range: Usd.PrimRange, predicate):
    """Convenience generator that prunes a prim range based on the given predicate"""
    for prim in prim_range:
        if predicate(prim):
            prim_range.PruneChildren()
        yield prim


def common_paths(paths: typing.Iterable[Sdf.Path]) -> typing.List[Sdf.Path]:
    """For the given paths, get those which are the common parents."""
    unique = list()
    for path in sorted(filter(lambda p: p and not p.IsAbsoluteRootPath(), paths)):
        if unique and path.HasPrefix(unique[-1]):  # we're a child, so safe to continue.
            continue
        unique.append(path)
    return unique


def iprims(stage: Usd.Stage, root_paths: typing.Iterable[Sdf.Path] = tuple(), prune_predicate: typing.Callable = None, traverse_predicate=Usd.PrimDefaultPredicate) -> typing.Iterator[Usd.Prim]:
    """Convenience function that creates a generator useful for common prim traversals.

    Without keyword arguments, this is the same as calling `Usd.Stage.Traverse(...)`, so
    use that instead when no `root_paths` or `prune_predicates` are needed.
    """
    if root_paths:  # Traverse only specific parts of the stage.
        root_paths = common_paths(root_paths)
        # Usd.PrimRange already discards invalid prims, so no need to check.
        root_prims = map(stage.GetPrimAtPath, root_paths)
        ranges = (Usd.PrimRange(prim, traverse_predicate) for prim in root_prims)
    else:
        ranges = [Usd.PrimRange.Stage(stage, traverse_predicate)]

    return chain.from_iterable(
        (_pruned_prims(iter(r), prune_predicate) for r in ranges) if prune_predicate else ranges
    )


@typing.overload
def edit_context(payload: Sdf.Payload, /, prim: Usd.Prim) -> Usd.EditContext:
    ...


@typing.overload
def edit_context(reference: Sdf.Reference, /, prim: Usd.Prim) -> Usd.EditContext:
    ...


@typing.overload
def edit_context(inherits: Usd.Inherits, /, path: Sdf.Path, layer: Sdf.Layer) -> Usd.EditContext:
    ...


@typing.overload
def edit_context(specializes: Usd.Specializes, /, path: Sdf.Path, layer: Sdf.Layer) -> Usd.EditContext:
    ...


@typing.overload
def edit_context(variant: Usd.VariantSet, /, layer: Sdf.Layer) -> Usd.EditContext:
    ...


@typing.overload
def edit_context(prim: Usd.Prim, /, query_filter: Usd.PrimCompositionQuery.Filter, arc_predicate: typing.Callable) -> Usd.EditContext:
    ...


@functools.singledispatch
def edit_context(obj, /, *args, **kwargs) -> Usd.EditContext:
    """Composition arcs target layer stacks. These functions help create EditTargets for the first matching node's root layer stack from prim's composition arcs.

    This allows for "chained" context switching while preserving the same stage objects.

    .. tip::

        You can try the below code snippet on ``USDView`` (or any other USD DCC application)
        Just swap the ``main = Usd.Stage.CreateInMemory()`` assignment for a stage on the viewport, e.g, for ``USDView``:

        .. code-block::  python

            >>> main = usdviewApi.stage

        Then paste the rest of the code as-is:

        .. image:: https://user-images.githubusercontent.com/8294116/133999486-b13e811a-91f4-4d8c-92d9-44c1f81b82d4.gif

    Example:
        >>> from pxr import Usd, UsdGeom, Sdf
        >>> main = Usd.Stage.CreateInMemory()
        >>> # Jump between 3 different layer stacks adding variants to the same set
        >>> # main [variant blue] -> reference [variant green] -> payload [variant red]
        >>> referenced = Usd.Stage.CreateInMemory()
        >>> referenced.SetDefaultPrim(referenced.DefinePrim("/Referenced"))
        >>> reference = Sdf.Reference(referenced.GetRootLayer().identifier)
        >>>
        >>> payloaded = Usd.Stage.CreateInMemory()
        >>> payloaded.SetDefaultPrim(payloaded.DefinePrim("/Payloaded"))
        >>> payload = Sdf.Payload(payloaded.GetRootLayer().identifier)
        >>>
        >>> top = main.DefinePrim("/Top")
        >>> top.GetReferences().AddReference(reference)
        True
        >>> import grill.usd as gusd
        >>> with gusd.edit_context(reference, top):
        ...     top.GetPayloads().AddPayload(payload)
        ...     with gusd.edit_context(payload, top):
        ...         geom = UsdGeom.Sphere.Define(main, top.GetPath().AppendPath("inner/child"))
        ...         color = geom.GetDisplayColorAttr()
        ...         color_set = geom.GetPrim().GetVariantSets().AddVariantSet("color")
        ...         color_set.AddVariant("from_payload")
        ...         color_set.SetVariantSelection("from_payload")
        ...         with gusd.edit_context(color_set, payloaded.GetRootLayer()):  # color_set.GetVariantEditContext() would fail here
        ...             color.Set([(1,0,0)])
        ...         color_set.ClearVariantSelection()
        ...     color_set.AddVariant("from_reference")
        ...     color_set.SetVariantSelection("from_reference")
        ...     with gusd.edit_context(color_set, referenced.GetRootLayer()):
        ...         color.Set([(0,1,0)])
        ...     color_set.ClearVariantSelection()
        ...
        True
        >>> color_set.AddVariant("from_top")
        >>> color_set.SetVariantSelection("from_top")
        >>> with color_set.GetVariantEditContext():
        ...     color.Set([(0,0,1)])
        ...
        >>> color_set.ClearVariantSelection()
        True
        >>> for each in main, referenced, payloaded:
        ...     print(each.GetRootLayer().ExportToString())
        ...
        #usda 1.0
        def "Top" (
            prepend references = @anon:0000019B6BE92A70:tmp.usda@
        )
        {
            over "inner"
            {
                over "child" (
                    prepend variantSets = "color"
                )
                {
                    variantSet "color" = {
                        "from_top" {
                            color3f[] primvars:displayColor = [(0, 0, 1)]
                        }
                    }
                }
            }
        }
        #usda 1.0
        (
            defaultPrim = "Referenced"
        )
        def "Referenced" (
            prepend payload = @anon:0000019B6BE93270:tmp.usda@
        )
        {
            over "inner"
            {
                over "child" (
                    prepend variantSets = "color"
                )
                {
                    variantSet "color" = {
                        "from_reference" {
                            color3f[] primvars:displayColor = [(0, 1, 0)]
                        }
                    }
                }
            }
        }
        #usda 1.0
        (
            defaultPrim = "Payloaded"
        )
        def "Payloaded"
        {
            def "inner"
            {
                def Sphere "child" (
                    prepend variantSets = "color"
                )
                {
                    variantSet "color" = {
                        "from_payload" {
                            color3f[] primvars:displayColor = [(1, 0, 0)]
                        }
                    }
                }
            }
        }

    """
    raise TypeError(f"Not implemented: {locals()}")  # lazy


@edit_context.register
def _(prim: Usd.Prim, /, query_filter, arc_predicate):
    # https://blogs.mathworks.com/developer/2015/03/31/dont-get-in-too-deep/
    # with write.context(prim, dict(kingdom="assets")):
    #     prim.GetAttribute("abc").Set(True)
    # with write.context(stage, dict(kingdom="category")):
    #     stage.DefinePrim("/hi")
    query = Usd.PrimCompositionQuery(prim)
    query.filter = query_filter
    for arc in query.GetCompositionArcs():
        if arc_predicate(arc):
            node = arc.GetTargetNode()
            target = Usd.EditTarget(node.layerStack.identifier.rootLayer, node)
            return Usd.EditContext(prim.GetStage(), target)
    raise ValueError(f"Could not find appropriate node for edit target for {prim} matching {arc_predicate}")


@edit_context.register(Sdf.Reference)
@edit_context.register(Sdf.Payload)
def _(arc: typing.Union[Sdf.Payload, Sdf.Reference], /, prim):
    identifier = arc.assetPath
    with Ar.ResolverContextBinder(prim.GetStage().GetPathResolverContext()):
        # Use Layer.Find since layer should have been open for the prim to exist.
        layer = Sdf.Layer.Find(identifier)
    if not layer:
        # Fallback to try find the layer directly. This might have been the result of an in memory stage.
        logger.debug(f"Layer with {identifier=} was not found on the resolver context for {prim=} at {prim.GetStage()}. Trying to find the layer outside of its context.")
        layer = Sdf.Layer.Find(identifier)
    if not layer:
        raise ValueError(f"Can't proceed without ability to find layer with {identifier=}")
    if not (arc.primPath or layer.defaultPrim):
        raise ValueError(f"Can't proceed without a prim path to target on arc {arc} for {layer}")
    path = arc.primPath or layer.GetPrimAtPath(layer.defaultPrim).path
    logger.debug(f"Searching to target {layer} on {path}")
    return _edit_context_by_arc(prim, type(arc), path, layer)


@edit_context.register(Usd.Inherits)
@edit_context.register(Usd.Specializes)
def _(arc_type: typing.Union[Usd.Inherits, Usd.Specializes], /, path, layer):
    return _edit_context_by_arc(arc_type.GetPrim(), type(arc_type), path, layer)


@edit_context.register
def _(variant_set: Usd.VariantSet, /, layer):
    with contextlib.suppress(Tf.ErrorException):
        return variant_set.GetVariantEditContext()
    # ----- From Pixar -----
    # pxr.Tf.ErrorException:
    # 	Error in '...::UsdVariantSet::GetVariantEditTarget' ...: 'Layer <identifier> is not a local layer of stage rooted at layer <identifier>'
    # https://graphics.pixar.com/usd/docs/api/class_usd_variant_set.html#a83f3adf614736a0b43fa1dd5271a9528
    # Currently, we require layer to be in the stage's local LayerStack (see UsdStage::HasLocalLayer()), and will issue an error and return an invalid EditTarget if layer is not.
    # We may relax this restriction in the future, if need arises, but it introduces several complications in specification and behavior.
    # ---------------------
    prim = variant_set.GetPrim()
    name = variant_set.GetName()
    selection = variant_set.GetVariantSelection()
    logger.debug(f"Searching target for {prim} with variant {name}, {selection} on {layer}")

    def is_target(arc):
        node = arc.GetTargetNode()
        return node.path.GetVariantSelection() == (name, selection) and layer == node.layerStack.identifier.rootLayer

    query_filter = Usd.PrimCompositionQuery.Filter()
    query_filter.arcTypeFilter = Usd.PrimCompositionQuery.ArcTypeFilter.Variant
    query_filter.hasSpecsFilter = Usd.PrimCompositionQuery.HasSpecsFilter.HasSpecs
    return edit_context(prim, query_filter, is_target)


def _edit_context_by_arc(prim, arc_type, path, layer):
    arc_filter = {
        Sdf.Payload: Usd.PrimCompositionQuery.ArcTypeFilter.Payload,
        Sdf.Reference: Usd.PrimCompositionQuery.ArcTypeFilter.Reference,
        Usd.Inherits: Usd.PrimCompositionQuery.ArcTypeFilter.Inherit,
        Usd.Specializes: Usd.PrimCompositionQuery.ArcTypeFilter.Specialize,
    }
    query_filter = Usd.PrimCompositionQuery.Filter()
    query_filter.arcTypeFilter = arc_filter[arc_type]

    def is_target(arc):
        node = arc.GetTargetNode()
        # USD-23.02 can do arc.GetTargetPrimPath() == path and arc.GetTargetLayer() == layer
        return node.path == path and node.layerStack.identifier.rootLayer == layer

    return edit_context(prim, query_filter, is_target)


@contextlib.contextmanager
def _prim_tree_printer(predicate, prims_to_include: typing.Container = frozenset()):
    prim_entry = Usd.Prim.GetName if predicate != Usd.PrimIsModel else lambda prim: f"{prim.GetName()} ({Usd.ModelAPI(prim).GetKind()})"

    class PrimTreePrinter(TreePrinter):
        """For everything else, use usdtree from the vanilla USD toolset"""

        def ftree(self, prim: Usd.Prim):
            self.ROOT = f"{super().ROOT}{prim_entry(prim)}"
            return super().ftree(prim)

    # another duck
    Usd.Prim.__iter__ = lambda prim: (p for p in prim.GetFilteredChildren(predicate) if not prims_to_include or p in prims_to_include)
    Usd.Prim.items = lambda prim: ((prim_entry(p), p) for p in prim)
    current = type(abc.Mapping).__instancecheck__  # can't unregister abc.Mapping.register, so use __instancecheck__

    type(abc.Mapping).__instancecheck__ = lambda cls, inst: current(cls, inst) or (cls == abc.Mapping and type(inst) == Usd.Prim)
    try:
        yield PrimTreePrinter()
    finally:
        type(abc.Mapping).__instancecheck__ = current
        del Usd.Prim.__iter__
        del Usd.Prim.items


def _format_prim_hierarchy(prims, include_descendants=True, predicate=Usd.PrimDefaultPredicate):
    for prim in prims:
        if prim.IsPseudoRoot():
            prims_to_tree = {prim}
            break
    else:
        root_paths = dict.fromkeys(common_paths((prim.GetPath() for prim in prims)))
        prims_to_tree = (prim for prim in prims if prim.GetPath() in root_paths)

    with _prim_tree_printer(predicate, set(prims) if not include_descendants else set()) as printer:
        return "\n".join(printer.ftree(prim) for prim in prims_to_tree)


# add other mesh creation utilities here?
def _make_plane(mesh, width, depth):
    # https://github.com/marcomusy/vedo/issues/86
    # https://blender.stackexchange.com/questions/230534/fastest-way-to-skin-a-grid
    x_ = np.linspace(-(width / 2), width / 2, width)
    z_ = np.linspace(depth / 2, - depth / 2, depth)
    X, Z = np.meshgrid(x_, z_)
    x = X.ravel()
    z = Z.ravel()
    y = np.zeros_like(x)
    points = np.stack((x, y, z), axis=1)
    xmax = x_.size
    zmax = z_.size
    faceVertexIndices = np.array([
        (i + j * xmax, i + j * xmax + 1, i + 1 + (j + 1) * xmax, i + (j + 1) * xmax)
        for j in range(zmax - 1) for i in range(xmax - 1)
    ])

    faceVertexCounts = np.full(len(faceVertexIndices), 4)
    with Sdf.ChangeBlock():
        mesh.GetPointsAttr().Set(points)
        mesh.GetFaceVertexCountsAttr().Set(faceVertexCounts)
        mesh.GetFaceVertexIndicesAttr().Set(faceVertexIndices)


class _GeomPrimvarInfo(enum.Enum):  # TODO: find a better name
    # One element for the entire Imageable prim; no interpolation.
    CONSTANT = UsdGeom.Tokens.constant, {UsdGeom.Imageable: 1}
    # One element for each face of the mesh; elements are typically not interpolated
    # but are inherited by other faces derived from a given face (via subdivision, tessellation, etc.).
    UNIFORM = UsdGeom.Tokens.uniform, {
        UsdGeom.Mesh: lambda mesh: len(mesh.GetFaceVertexCountsAttr().Get()),
        UsdGeom.BasisCurves: lambda curve: curve.ComputeUniformDataSize(0),
        UsdGeom.Sphere: 100,  # TODO: there must be a better way of finding these numbers.
        UsdGeom.Cube: 6,
        UsdGeom.Capsule: 90,
        UsdGeom.Cone: 20,
        UsdGeom.Cylinder: 30,
    }
    # One element for each point of the mesh; interpolation of point data is:
    #   Varying: always linear.
    VARYING = UsdGeom.Tokens.varying, {
        UsdGeom.Mesh: lambda mesh: len(mesh.GetPointsAttr().Get()),
        UsdGeom.BasisCurves: lambda curve: curve.ComputeVaryingDataSize(0),
        UsdGeom.Sphere: 92,
        UsdGeom.Cube: 8,
        UsdGeom.Capsule: 82,
        UsdGeom.Cone: 31,
        UsdGeom.Cylinder: 42,
    }
    #   Vertex: applied according to the subdivisionScheme attribute.
    VERTEX = UsdGeom.Tokens.vertex, {**VARYING[1], UsdGeom.BasisCurves: lambda curve: curve.ComputeVertexDataSize(0)}
    # One element for each of the face-vertices that define the mesh topology;
    # interpolation of face-vertex data may be smooth or linear, according to the
    # subdivisionScheme and faceVaryingLinearInterpolation attributes.
    FACE_VARYING = UsdGeom.Tokens.faceVarying, {
        UsdGeom.Mesh: lambda mesh: len(mesh.GetFaceVertexIndicesAttr().Get()),
        UsdGeom.Sphere: 380,
        UsdGeom.Cube: 24,
        UsdGeom.Capsule: 340,
        UsdGeom.Cone: 70,
        UsdGeom.Cylinder: 100,
    }

    def size(self, prim):
        for geom_class, value in self.value[1].items():
            if geom := geom_class(prim):
                return value(geom) if callable(value) else value
        raise TypeError(f"Don't know how to compute '{self.interpolation()}' size on {self} for prim type '{prim.GetPrim().GetTypeName()}': {prim}")

    def interpolation(self):
        return self.value[0]



from functools import partial, cache
from pxr import UsdUtils

from ._qt import QtWidgets, QtCore, QtGui


@cache
def _report_no_charts():
    print("No QtCharts module could be imported. Are you in a DCC app?")


class _ContainerTree(QtWidgets.QTreeWidget):
    def __init__(self, *args, value=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.setColumnCount(2)  # 0: key, 1: value
        self.setHeaderHidden(True)
        self.setSelectionMode(self.SelectionMode.ExtendedSelection)
        self.setData(value)

    def setData(self, value):
        self.clear()
        containers = (dict, list, tuple)
        def populate(parent, data):
            if isinstance(data, containers):
                for _key, _value in data.items() if isinstance(data, dict) else enumerate(data):
                    display = None if isinstance(_value, containers) else str(_value)
                    populate(QtWidgets.QTreeWidgetItem(parent, [str(_key), display]), _value)

        populate(self.invisibleRootItem(), value)
        self.expandAll()
        self.resizeColumnToContents(0)
        self.resizeColumnToContents(1)


class _StatsPie(QtWidgets.QWidget):

    def setStats(self, value, *, title=None):
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        def populate(parent, data: dict, chart_title: str = ""):
            def show_label(pie, state):
                pie.setExploded(state)

            substats = []
            stats = []
            for data_key, data_value  in data.items():
                target = substats if isinstance(data_value, dict) else stats
                target.append((data_key, data_value))

            def _add_pie():
                try:
                    from ._qt import QtCharts  # Hou-19.5 & Maya-2023 don't include QtCharts
                except ImportError:
                    _report_no_charts()
                    return
                series = QtCharts.QPieSeries()
                for stat_key, stat_value in stats:
                    series.append(f"{stat_key}: {stat_value}", stat_value)
                series.setLabelsVisible(True)
                for slice in series.slices():
                    slice.hovered.connect(partial(show_label, slice))
                chart = QtCharts.QChart()
                chart.addSeries(series)
                chart.legend().hide()

                if chart_title:
                    chart.setTitle(chart_title)
                view = QtCharts.QChartView(chart)
                view.setRenderHint(QtGui.QPainter.Antialiasing)
                parent.addWidget(view)

            if stats:
                _add_pie()

            if substats:
                subparent = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
                parent.addWidget(subparent)
                for substat_key, substat_value in substats:
                    subtitle = f"{'' if stats else chart_title + ': '}{substat_key}"
                    populate(subparent, substat_value, chart_title=subtitle)  # lazy

        main_splitter = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        layout.addWidget(main_splitter)
        # Note: this is relying on current structure from return value of
        # UsdUtils.ComputeUsdStageStats dict[str: (str|int|dict)]
        # if that changes we might have to change, but for now don't think much about it
        populate(main_splitter, value, chart_title=title)
        self.setLayout(layout)


class StageStats(QtWidgets.QDialog):
    def __init__(self, *args, stage=None, **kwargs):
        super().__init__(*args, **kwargs)
        layout = QtWidgets.QHBoxLayout()
        self._usd_tree = _ContainerTree(parent=self)
        self._pie = _StatsPie(parent=self)
        splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        splitter.addWidget(self._pie)
        splitter.addWidget(self._usd_tree)
        layout.addWidget(splitter)
        self.setLayout(layout)
        if stage:
            self.setStage(stage)

    def setStage(self, stage):
        stats = UsdUtils.ComputeUsdStageStats(stage)
        self._usd_tree.setData(stats)
        self._pie.setStats(stats, title=f'Stats for {stage.GetRootLayer().identifier}')



import os
import io
import csv
import shutil
import tempfile
import unittest
from unittest import mock

from pxr import Usd, UsdGeom, Sdf, UsdShade

from grill import cook, usd, names
from grill.views import description, sheets, create, _attributes, stats, _core, _graph, _qt
from grill.views._qt import QtWidgets, QtCore, QtGui

# 2024-02-03 - Python-3.12 & USD-23.11
# leaving the PySide6 import below freezes windows in Python-3.12. Importing it first when running tests "fixes" the freeze.
# from PySide6 import QtWebEngineWidgets
# alternatively, the following can be called to ensure shared contexts are set, which also prevent the freeze of the application:
QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
# but don't want to use that since that needs to be set prior to an application initialization (which grill can't control as in USDView, Maya, Houdini...)
# https://stackoverflow.com/questions/56159475/qt-webengine-seems-to-be-initialized

# 2024-02-03
# python -m unittest --durations 0 test_views
# Slowest test durations
# ----------------------------------------------------------------------
# 1.963s     test_scenegraph_composition (test_views.TestViews.test_scenegraph_composition)
# 1.882s     test_taxonomy_editor (test_views.TestViews.test_taxonomy_editor)
# 1.579s     test_content_browser (test_views.TestViews.test_content_browser)
# 0.789s     test_spreadsheet_editor (test_views.TestViews.test_spreadsheet_editor)
# 0.383s     test_horizontal_scroll (test_views.TestGraphicsViewport.test_horizontal_scroll)
# 0.329s     test_connection_view (test_views.TestViews.test_connection_view)
# 0.322s     test_layer_stack_hovers (test_views.TestViews.test_layer_stack_hovers)
# 0.204s     test_dot_call (test_views.TestViews.test_dot_call)
# 0.169s     test_display_color_editor (test_views.TestViews.test_display_color_editor)
# 0.167s     test_stats (test_views.TestViews.test_stats)
# 0.121s     test_prim_filter_data (test_views.TestViews.test_prim_filter_data)
# 0.116s     test_prim_composition (test_views.TestViews.test_prim_composition)
# 0.106s     test_create_assets (test_views.TestViews.test_create_assets)
# 0.014s     test_pan (test_views.TestGraphicsViewport.test_pan)
#
# (durations < 0.001s were hidden; use -v to show these durations)
# ----------------------------------------------------------------------
# Ran 18 tests in 8.216s


class TestPrivate(unittest.TestCase):
    def test_common_paths(self):
        input_paths = [
            Sdf.Path("/world/hi"),
            Sdf.Path.absoluteRootPath,
            Sdf.Path("/hola/hello/new1"),
            Sdf.Path("/world/child/nested"),
            Sdf.Path("/invalid/1"),
            Sdf.Path("/hola/hello/new2"),
            Sdf.Path("/hola/hello/new2/twochild"),
            Sdf.Path("/hola/hello/new2/twochild/more"),
            Sdf.Path("/hola/hello/new2/a"),
            Sdf.Path("/hola/hello/new2/zzzzzzzzzzzzzzzzzzz"),
            Sdf.Path("/hola/hello/new3"),
            Sdf.Path("/hola/hello/n9/nested/one"),
            Sdf.Path("/hola/hello/new01/nested/deep"),
            Sdf.Path("/hola/hello/n9/nested/two"),
            Sdf.Path("/hola/bye/child"),
            Sdf.Path("/deep/nested/unique/path"),
            Sdf.Path("/alone"),
        ]
        actual = usd.common_paths(input_paths)
        expected = [
            Sdf.Path('/alone'),
            Sdf.Path('/deep/nested/unique/path'),
            Sdf.Path('/hola/bye/child'),
            Sdf.Path('/hola/hello/n9/nested/one'),
            Sdf.Path('/hola/hello/n9/nested/two'),
            Sdf.Path('/hola/hello/new01/nested/deep'),
            Sdf.Path('/hola/hello/new1'),
            Sdf.Path('/hola/hello/new2'),
            Sdf.Path('/hola/hello/new3'),
            Sdf.Path('/world/child/nested'),
            Sdf.Path('/world/hi'),
        ]
        self.assertEqual(actual, expected)

    def test_core(self):
        _core._ensure_dot()


class TestViews(unittest.TestCase):
    def setUp(self):
        self._app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])

        sphere = Usd.Stage.CreateInMemory()
        UsdGeom.Sphere.Define(sphere, "/sph")
        root_path = "/root"
        sphere_root = sphere.DefinePrim(root_path)
        sphere_root.CreateAttribute("greet", Sdf.ValueTypeNames.String).Set("hello")
        sphere.SetDefaultPrim(sphere_root)

        capsule = Usd.Stage.CreateInMemory()
        UsdGeom.Capsule.Define(capsule, "/cap")
        root_path = "/root"
        capsule_root = capsule.DefinePrim(root_path)
        capsule_root.CreateAttribute("who", Sdf.ValueTypeNames.String).Set("world")
        capsule.SetDefaultPrim(capsule_root)

        merge = Usd.Stage.CreateInMemory()
        for i in (capsule, sphere):
            merge.GetRootLayer().subLayerPaths.append(i.GetRootLayer().identifier)
        merge.SetDefaultPrim(merge.GetPrimAtPath(root_path))

        world = Usd.Stage.CreateInMemory()
        self.nested = world.DefinePrim("/nested/child")
        self.sibling = world.DefinePrim("/nested/sibling")
        self.nested.GetReferences().AddReference(merge.GetRootLayer().identifier)

        self.capsule = capsule
        self.sphere = sphere
        self.merge = merge
        self.world = world

        self._tmpf = tempfile.mkdtemp()
        self._token = cook.Repository.set(cook.Path(self._tmpf) / "repo")
        self.rootf = names.UsdAsset.get_anonymous()
        self.grill_world = gworld = cook.fetch_stage(self.rootf.name)
        self.person = cook.define_taxon(gworld, "Person")
        self.agent = cook.define_taxon(gworld, "Agent", references=(self.person,))
        self.generic_agent = cook.create_unit(self.agent, "GenericAgent")

    def tearDown(self) -> None:
        cook.Repository.reset(self._token)
        # Reset all members to USD objects to ensure the used layers are cleared
        # (otherwise in Windows this can cause failure to remove the temporary files)
        self.generic_agent = None
        self.agent = None
        self.person = None
        self.grill_world = None
        self.capsule = None
        self.sphere = None
        self.merge = None
        self.world = None
        self.nested = None
        self.sibling = None
        shutil.rmtree(self._tmpf)
        self._app.quit()

    def test_connection_view(self):
        for graph_viewer in _graph.GraphView, _graph._GraphSVGViewer:
            with self.subTest(graph_viewer=graph_viewer):
                _graph._GraphViewer = graph_viewer
                if graph_viewer == _graph._GraphSVGViewer:
                    for pixmap_enabled in True, False:
                        with self.subTest(pixmap_enabled=pixmap_enabled):
                            _graph._USE_SVG_VIEWPORT = pixmap_enabled
                            self._sub_test_connection_view()
                else:
                    self._sub_test_connection_view()

    def _sub_test_connection_view(self):
        # https://openusd.org/release/tut_simple_shading.html
        stage = Usd.Stage.CreateInMemory()
        material = UsdShade.Material.Define(stage, '/TexModel/boardMat')
        pbrShader = UsdShade.Shader.Define(stage, '/TexModel/boardMat/PBRShader')
        pbrShader.CreateInput("roughness", Sdf.ValueTypeNames.Float).Set(0.4)
        material.CreateSurfaceOutput().ConnectToSource(pbrShader.ConnectableAPI(), "surface")
        # Ensure cycles don't cause recursion
        cycle_input = pbrShader.CreateInput("cycle_in", Sdf.ValueTypeNames.Float)
        cycle_output = pbrShader.CreateOutput("cycle_out", Sdf.ValueTypeNames.Float)
        cycle_output.ConnectToSource(cycle_input)
        description._graph_from_connections(material)
        viewer = description._ConnectableAPIViewer()
        viewer.setPrim(material)
        viewer.setPrim(None)

    def test_scenegraph_composition(self):
        for graph_viewer in _graph.GraphView, _graph._GraphSVGViewer:
            with self.subTest(graph_viewer=graph_viewer):
                _graph._GraphViewer = graph_viewer
                if graph_viewer == _graph._GraphSVGViewer:
                    for pixmap_enabled in True, False:
                        with self.subTest(pixmap_enabled=pixmap_enabled):
                            _graph._USE_SVG_VIEWPORT = pixmap_enabled
                            self._sub_test_scenegraph_composition()
                            self._sub_test_layer_stack_bidirectionality()
                else:
                    self._sub_test_scenegraph_composition()
                    self._sub_test_layer_stack_bidirectionality()

    def _sub_test_scenegraph_composition(self):
        widget = description.LayerStackComposition()
        widget.setStage(self.world)

        # cheap. All these layers affect a single prim
        affectedPaths = dict.fromkeys((i.GetRootLayer() for i in (self.capsule, self.sphere, self.merge)), 1)

        # the world affects both root and the nested prims, stage layer stack is included
        affectedPaths.update(dict.fromkeys(self.world.GetLayerStack(), 3))

        for row in range(widget._layers.model.rowCount()):
            layer = widget._layers.model._objects[row]
            widget._layers.table.selectRow(row)
            expectedAffectedPrims = affectedPaths[layer]
            actualListedPrims = widget._prims.model.rowCount()
            self.assertEqual(expectedAffectedPrims, actualListedPrims)

        widget._layers.table.selectAll()
        self.assertEqual(len(affectedPaths), widget._layers.model.rowCount())
        self.assertEqual(3, widget._prims.model.rowCount())

        widget.setPrimPaths({"/nested/sibling"})
        widget.setStage(self.world)

        widget._layers.table.selectAll()
        self.assertEqual(2, widget._layers.model.rowCount())
        self.assertEqual(1, widget._prims.model.rowCount())

        widget._has_specs.setChecked(True)
        widget._graph_edge_include[description.Pcp.ArcTypeReference].setChecked(False)
        # add_dll_directory only on Windows
        os.add_dll_directory = lambda path: print(f"Added {path}") if not hasattr(os, "add_dll_directory") else os.add_dll_directory

        _core._which.cache_clear()
        with mock.patch("grill.views.description._which") as patch:  # simulate dot is not in the environment
            patch.return_value = None
            widget._graph_view.view([0,1])

        _core._which.cache_clear()
        with mock.patch("grill.views.description.nx.nx_agraph.write_dot") as patch:  # simulate pygraphviz is not installed
            patch.side_effect = ImportError
            widget._graph_view.view([0])

        widget.deleteLater()

    def _sub_test_layer_stack_bidirectionality(self):
        """Confirm that bidirectionality between layer stacks completes.

        Bidirectionality in the composition graph is achieved by:
            - parent_stage -> child_stage via a reference, payload arcs
            - child_stage -> parent_stage via a inherits, specializes arcs
        """
        parent_stage = Usd.Stage.CreateInMemory()
        child_stage = Usd.Stage.CreateInMemory()
        prim = parent_stage.DefinePrim("/a/b")
        child_prim = child_stage.DefinePrim("/child")
        child_prim.GetInherits().AddInherit("/foo")
        child_prim.GetSpecializes().AddSpecialize("/foo")
        child_stage.SetDefaultPrim(child_prim)
        child_identifier = child_stage.GetRootLayer().identifier
        prim.GetReferences().AddReference(child_identifier)
        prim.GetPayloads().AddPayload(child_identifier)

        widget = description.LayerStackComposition()
        widget.setStage(parent_stage)
        widget._layers.table.selectAll()

        graph_view = widget._graph_view

    def test_layer_stack_hovers(self):
        _graph._GraphViewer = _graph.GraphView
        _graph._USE_SVG_VIEWPORT = False

        parent_stage = Usd.Stage.CreateInMemory()
        child_stage = Usd.Stage.CreateInMemory()
        prim = parent_stage.DefinePrim("/a/b")
        child_prim = child_stage.DefinePrim("/child")
        child_prim.GetInherits().AddInherit("/foo")
        child_prim.GetSpecializes().AddSpecialize("/foo")
        child_stage.SetDefaultPrim(child_prim)
        child_identifier = child_stage.GetRootLayer().identifier
        prim.GetReferences().AddReference(child_identifier)
        prim.GetPayloads().AddPayload(child_identifier)

        widget = description.LayerStackComposition()
        widget.setStage(parent_stage)
        widget._graph_precise_source_ports.setChecked(True)
        widget._has_specs.setCheckState(QtCore.Qt.CheckState.PartiallyChecked)

        widget._layers.table.selectAll()
        graph_view = widget._graph_view
        cycle_collected = False
        nodes_hovered_checked = False
        for item in graph_view.scene().items():
            item.boundingRect()  # trigger bounding rect logic
            if isinstance(item, _graph._Edge):
                cycle_collected = True
            if isinstance(item, _graph._Node) and item.isVisible():
                nodes_hovered_checked = True

                # Test hover with no modifiers
                event = QtWidgets.QGraphicsSceneHoverEvent(QtCore.QEvent.GraphicsSceneHoverMove)
                center = item.sceneBoundingRect().center()
                event.setScenePos(center)
                item.hoverEnterEvent(event)
                self.assertEqual(item.cursor().shape(), QtGui.Qt.ArrowCursor)
                self.assertEqual(item.textInteractionFlags(), item._default_text_interaction)
                item.hoverLeaveEvent(event)

                # Test hover with Ctrl modifier
                event = QtWidgets.QGraphicsSceneHoverEvent(QtCore.QEvent.GraphicsSceneHoverMove)
                event.setScenePos(center)
                event.setModifiers(QtCore.Qt.ControlModifier)
                item.hoverEnterEvent(event)
                self.assertEqual(item.cursor().shape(), QtGui.Qt.PointingHandCursor)
                item.hoverLeaveEvent(event)

                # Test hover with Alt modifier
                event = QtWidgets.QGraphicsSceneHoverEvent(QtCore.QEvent.GraphicsSceneHoverMove)
                event.setScenePos(item.sceneBoundingRect().center())
                event.setModifiers(QtCore.Qt.AltModifier)
                item.hoverEnterEvent(event)
                self.assertEqual(item.cursor().shape(), QtGui.Qt.ClosedHandCursor)
                self.assertEqual(item.textInteractionFlags(), QtCore.Qt.NoTextInteraction)
                item.hoverLeaveEvent(event)

        self.assertTrue(cycle_collected)
        self.assertTrue(nodes_hovered_checked)

    def test_prim_composition(self):
        for pixmap_enabled in True, False:
            with self.subTest(pixmap_enabled=pixmap_enabled):
                description._SVG_AS_PIXMAP = pixmap_enabled
                self._sub_test_prim_composition()

    def _sub_test_prim_composition(self):
        temp = Usd.Stage.CreateInMemory()
        temp.GetRootLayer().subLayerPaths = [self.nested.GetStage().GetRootLayer().identifier]
        prim = temp.GetPrimAtPath(self.nested.GetPath())
        widget = description.PrimComposition()
        widget.setPrim(prim)

        # cheap. prim is affected by 2 layers
        # single child for this prim.
        self.assertTrue(widget.composition_tree._model.invisibleRootItem().hasChildren())

        widget._complete_target_layerstack.setChecked(True)
        widget.setPrim(prim)
        self.assertTrue(widget.composition_tree._model.invisibleRootItem().hasChildren())

        with mock.patch("grill.views.description.QtWidgets.QApplication.keyboardModifiers") as patch:
            patch.return_value = QtCore.Qt.ShiftModifier
            root_idx = widget.composition_tree._model.index(0, 0)
            widget.composition_tree.expanded.emit(root_idx)
            widget.composition_tree.collapsed.emit(root_idx)

        widget.setPrim(None)
        self.assertFalse(widget.composition_tree._model.invisibleRootItem().hasChildren())

        widget.clear()

    def test_create_assets(self):
        stage = cook.fetch_stage(str(self.rootf))

        for each in range(1, 6):
            cook.define_taxon(stage, f"Option{each}")

        widget = create.CreateAssets()
        widget.setStage(stage)

        widget._amount.setValue(3)  # TODO: create 10 assets, clear tmp directory

        data = (
            ['Option1', 'asset01', 'Asset 01', 'Description 01'],
            ['Option2', 'asset02', 'Asset 02', 'Description 02'],
            ['Option2', '',        'Asset 03', 'Description 03'],
        )

        QtWidgets.QApplication.instance().clipboard().setText('')
        widget.sheet._pasteClipboard()

        stream = io.StringIO()
        csv.writer(stream, delimiter=csv.excel_tab.delimiter).writerows(data)
        QtWidgets.QApplication.instance().clipboard().setText(stream.getvalue())

        widget.sheet.table.selectAll()
        widget.sheet._pasteClipboard()
        widget._create()
        taxon_editor = widget.sheet._columns[0].editor(widget, None, None)
        self.assertIsInstance(taxon_editor, QtWidgets.QComboBox)
        widget._apply()

    def test_taxonomy_editor(self):
        for graph_viewer in _graph.GraphView, _graph._GraphSVGViewer:
            with self.subTest(graph_viewer=graph_viewer):
                _graph._GraphViewer = graph_viewer
                if graph_viewer == _graph._GraphSVGViewer:
                    for pixmap_enabled in True, False:
                        with self.subTest(pixmap_enabled=pixmap_enabled):
                            _graph._USE_SVG_VIEWPORT = pixmap_enabled
                            self._sub_test_taxonomy_editor()
                else:
                    self._sub_test_taxonomy_editor()

    def _sub_test_taxonomy_editor(self):
        stage = cook.fetch_stage(str(self.rootf.get_anonymous()))

        existing = [cook.define_taxon(stage, f"Option{each}") for each in range(1, 6)]
        widget = create.TaxonomyEditor()
        if isinstance(widget._graph_view, _graph.GraphView):
            with self.assertRaisesRegex(LookupError, "Could not find sender"):
                invalid_uril = QtCore.QUrl(f"{widget._graph_view.url_id_prefix}not_a_digit")
                widget._graph_view._graph_url_changed(invalid_uril)
        else:
            with self.assertRaisesRegex(RuntimeError, "'graph' attribute not set yet"):
                invalid_uril = QtCore.QUrl(f"{widget._graph_view.url_id_prefix}not_a_digit")
                widget._graph_view._graph_url_changed(invalid_uril)
        widget.setStage(stage)

        widget._amount.setValue(3)  # TODO: create 10 assets, clear tmp directory

        valid_data = (
            ['NewType1', 'Option1', 'Id1', ],
            ['NewType2', '', 'Id2', ],
        )
        data = valid_data + (
            ['',         'Option1', 'Id3', ],
        )

        QtWidgets.QApplication.instance().clipboard().setText('')
        widget.sheet._pasteClipboard()

        stream = io.StringIO()
        csv.writer(stream, delimiter=csv.excel_tab.delimiter).writerows(data)
        QtWidgets.QApplication.instance().clipboard().setText(stream.getvalue())

        widget.sheet.table.selectAll()
        widget.sheet._pasteClipboard()
        widget._create()

        for name, __, __ in valid_data:
            created = stage.GetPrimAtPath(cook._TAXONOMY_ROOT_PATH).GetPrimAtPath(name)
            self.assertTrue(created.IsValid())

        sheet_model = widget.sheet.model
        index = sheet_model.index(0, 1)
        editor = widget.sheet._columns[1].editor(None, None, index)
        self.assertIsInstance(editor, QtWidgets.QDialog)
        widget.sheet._columns[1].setter(editor, sheet_model, index)
        editor._options.selectAll()
        menu = editor._create_context_menu()
        menu.actions()[0].trigger()

        # after creation, set stage again to test existing column
        widget._apply()
        widget._existing.table.selectAll()
        selected_items = widget._existing.table.selectedIndexes()
        self.assertEqual(len(selected_items), len(valid_data) + len(existing))

        if isinstance(widget._graph_view, _graph.GraphView):
            sender = next(iter(widget._graph_view._nodes_map.values()), None)
            self.assertIsNotNone(sender, msg=f"Expected sender to be an actual object of type {_graph._Node}. Got None, check pygraphviz / pydot requirements")
            sender.linkActivated.emit("")
        else:
            valid_url = QtCore.QUrl(f"{widget._graph_view.url_id_prefix}{existing[-1].GetName()}")
            widget._graph_view._graph_url_changed(valid_url)
            # Nitpick, wait for dot 2 svg conversions to finish
            # This does not crash the program but an exception is logged when race
            # conditions apply (e.g. the object is deleted before the runnable completes).
            # This logged exception comes in the form of:
            # RuntimeError: Internal C++ object (_Dot2SvgSignals) already deleted.
            # Solution seems to be to block and wait for all runnables to complete.
            widget._graph_view._threadpool.waitForDone(10_000)

    def test_spreadsheet_editor(self):
        widget = sheets.SpreadsheetEditor()
        widget._model_hierarchy.setChecked(False)  # default is True
        self.world.OverridePrim("/child_orphaned")
        self.nested.SetInstanceable(True)
        widget._orphaned.setChecked(True)
        assert self.nested.IsInstance()
        widget.setStage(self.world)
        self.assertEqual(self.world, widget.stage)
        widget.table.scrollContentsBy(10, 10)

        widget.table.selectAll()
        expected_rows = {0, 1, 2, 3}  # 3 prims from path: /nested, /nested/child, /nested/sibling, /child_orphaned
        visible_rows = ({i.row() for i in widget.table.selectedIndexes()})
        self.assertEqual(expected_rows, visible_rows)

        widget.table.clearSelection()
        widget._column_options[0]._line_filter.setText("chi")
        widget._column_options[0]._updateMask()
        widget.table.resizeColumnToContents(0)

        widget.table.selectAll()
        expected_rows = {0, 1}  # 1 prim from filtered name: /nested/child
        visible_rows = ({i.row() for i in widget.table.selectedIndexes()})
        self.assertEqual(expected_rows, visible_rows)

        widget._copySelection()
        clip = QtWidgets.QApplication.instance().clipboard().text()
        data = tuple(csv.reader(io.StringIO(clip), delimiter=csv.excel_tab.delimiter))
        expected_data = (
            ['/nested/child', 'child', '', '', '', 'True', '', 'False'],
            ['/child_orphaned', 'child_orphaned', '', '', '', 'False', '', 'False'],
        )
        self.assertEqual(data, expected_data)

        widget.table.clearSelection()

        widget._model_hierarchy.click()  # enables model hierarchy, which we don't have any
        widget.table.selectAll()
        expected_rows = set()
        visible_rows = ({i.row() for i in widget.table.selectedIndexes()})
        self.assertEqual(expected_rows, visible_rows)

        widget.table.clearSelection()

        widget._lock_all.click()
        widget._conformLockSwitch()
        widget._vis_all.click()
        widget._conformVisibilitySwitch()

        widget._column_options[0]._line_filter.setText("")
        widget._model_hierarchy.click()  # disables model hierarchy, which we don't have any
        widget.table.selectAll()
        _log = lambda *args: print(args)
        with mock.patch(f"{QtWidgets.__name__}.QMessageBox.warning", new=_log):
            widget._pasteClipboard()

        widget.model._prune_children = {Sdf.Path("/pruned")}
        gworld = self.grill_world
        with cook.unit_context(self.generic_agent):
            child_agent = gworld.DefinePrim(self.generic_agent.GetPath().AppendChild("child"))
            child_attr = child_agent.CreateAttribute("agent_greet", Sdf.ValueTypeNames.String, custom=False)
            child_attr.Set("aloha")
        agent_id = cook.unit_asset(self.generic_agent)
        for i in range(3):
            agent = gworld.DefinePrim(f"/Instanced/Agent{i}")
            agent.GetReferences().AddReference(agent_id.identifier)
            agent.SetInstanceable(True)
        agent.SetActive(False)
        gworld.OverridePrim("/non/existing/prim")
        gworld.DefinePrim("/pruned/prim")
        inactive = gworld.DefinePrim("/another_inactive")
        inactive.SetActive(False)
        gworld.GetRootLayer().subLayerPaths.append(self.world.GetRootLayer().identifier)
        widget._column_options[0]._line_filter.setText("")
        widget.table.clearSelection()
        widget._active.setChecked(False)
        widget._classes.setChecked(True)
        widget._filters_logical_op.setCurrentIndex(1)
        widget.stage = gworld
        widget.table.selectAll()
        expected_colors = {str(each.value): each for each in sheets._PrimTextColor}  # colors are not hashable
        expected_fonts = {each.weight() for each in (  # font not hashable in PySide2
            sheets._prim_font(),
            sheets._prim_font(abstract=True),
            sheets._prim_font(abstract=True, orphaned=True),
            sheets._prim_font(orphaned=True),
        )}
        self.assertEqual(len(expected_fonts), 3)  # three weights: Light, ExtraLight, Normal
        collected_fonts = set()
        for each in widget.table.selectionModel().selectedIndexes():
            color_key = str(each.data(role=QtCore.Qt.ForegroundRole))
            font = each.data(role=QtCore.Qt.FontRole)
            font_key = font.weight()
            expected_colors.pop(color_key, None)
            collected_fonts.add(font_key)

        self.assertEqual(expected_colors, dict())
        self.assertEqual(expected_fonts, collected_fonts)

    def test_prim_filter_data(self):
        stage = cook.fetch_stage(self.rootf)
        person = cook.define_taxon(stage, "Person")
        agent = cook.define_taxon(stage, "Agent", references=(person,))
        generic = cook.create_unit(agent, "GenericAgent")
        with cook.unit_context(generic):
            stage.DefinePrim(generic.GetPath().AppendChild("ChildPrim"))
        generic_asset = cook.unit_asset(generic)
        for each in range(10):
            path = Sdf.Path.absoluteRootPath.AppendChild(f"Instance{each}")
            instance = stage.DefinePrim(path)
            instance.SetInstanceable(True)
            instance.GetPayloads().AddPayload(generic_asset.identifier)
        instance.SetActive(False)
        instance.Unload()
        over = stage.OverridePrim("/Orphaned")
        widget = sheets.SpreadsheetEditor()
        for stage_value in (stage, None):
            widget.setStage(stage_value)
            for each in range(2):
                widget._filters_logical_op.setCurrentIndex(each)
                widget._model_hierarchy.click()  # default is True
                widget._orphaned.click()
                widget._classes.click()
                widget._defined.click()
                widget._active.click()
                widget._inactive.click()
        widget.model._root_paths = {over.GetPath(), over.GetPath()}
        widget.model._prune_children = {over.GetPath()}
        widget.setStage(stage)

    def test_dot_call(self):
        """Test execution of function by mocking dot with python call"""
        with mock.patch("grill.views.description._which") as patch:
            patch.return_value = 'python'
            error, targetpath = _graph._dot_2_svg('nonexisting_path')
            # an error would be reported back
            self.assertIsNotNone(error)

    def test_content_browser(self):
        stage = cook.fetch_stage(self.rootf)
        taxon = cook.define_taxon(stage, "Another")
        parent, child = cook.create_many(taxon, ['A', 'B'])
        for path, value in (
                ("", (2, 15, 6)),
                ("Deeper/Nested/Golden1", (-4, 5, 1)),
                ("Deeper/Nested/Golden2", (-4, -10, 1)),
                ("Deeper/Nested/Golden3", (0, 10, -2)),
        ):
            spawned = UsdGeom.Xform(cook.spawn_unit(parent, child, path))
            spawned.AddTranslateOp().Set(value=value)
        variant_set_name = "testset"
        variant_name = "testvar"
        vset = child.GetVariantSet(variant_set_name)
        vset.AddVariant(variant_name)
        vset.SetVariantSelection(variant_name)
        with vset.GetVariantEditContext():
            stage.DefinePrim(child.GetPath().AppendChild("in_variant"))
        path_with_variant = child.GetPath().AppendVariantSelection(variant_set_name, variant_name)

        layers = stage.GetLayerStack()
        args = stage.GetLayerStack(), None, stage.GetPathResolverContext(), (Sdf.Path("/"), spawned.GetPrim().GetPath(), path_with_variant)
        anchor = layers[0]

        def _log(*args):
            print(args)

        _core_run = _core._run

        def _fake_run(run_args: list):
            return "", Sdf.Layer.FindOrOpen(run_args[-1]).ExportToString()

        # sdffilter still not coming via pypi, so patch for now
        with mock.patch("grill.views.description._core._run", new=_fake_run if not description._which("sdffilter") else _core_run):
            dialog = description._start_content_browser(*args)
            browser: description._PseudoUSDBrowser = dialog.findChild(description._PseudoUSDBrowser)
            assert browser._browsers_by_layer.values()
            first_browser_widget, = browser._browsers_by_layer.values()
            first_browser_widget._format_options.setCurrentIndex(0)  # pseudoLayer (through sdffilter)
            first_browser_widget._format_options.setCurrentIndex(1)  # outline (through sdffilter)
            first_browser_widget._format_options.setCurrentIndex(2)  # usdtree (through usdtree)
            first_browser_widget._format_options.setCurrentIndex(0)

            browser_tab: description._PseudoUSDTabBrowser = first_browser_widget.findChild(description._PseudoUSDTabBrowser)
            browser._on_identifier_requested(anchor, layers[1].identifier)
            with mock.patch(f"{QtWidgets.__name__}.QMessageBox.warning", new=_log):
                browser._on_identifier_requested(anchor, "/missing/file.usd")
                _, empty_png = tempfile.mkstemp(suffix=".png")
                browser._on_identifier_requested(anchor, empty_png)
                _, empty_usd = tempfile.mkstemp(suffix=".usda")
                browser._on_identifier_requested(anchor, empty_usd)

            menu = browser._menu_for_tab(0)
            self.assertTrue(bool(menu.actions()))

            position = QtCore.QPoint(10, 10)
            pixelDelta = QtCore.QPoint(0, 0)
            angleDelta_zoomIn = QtCore.QPoint(0, 120)
            buttons = QtCore.Qt.NoButton
            modifiers = QtCore.Qt.ControlModifier
            phase = QtCore.Qt.NoScrollPhase
            inverted = False

            # ZOOM IN
            event = QtGui.QWheelEvent(position, position, pixelDelta, angleDelta_zoomIn, buttons, modifiers, phase, inverted)
            browser_tab.wheelEvent(event)

            # Assert that the scale has changed according to the zoom logic
            angleDelta_zoomOut = QtCore.QPoint(-120, 0)

            # ZOOM OUT
            event = QtGui.QWheelEvent(position, position, pixelDelta, angleDelta_zoomOut, buttons, modifiers, phase, inverted)
            browser_tab.wheelEvent(event)

            browser._close_many(range(len(browser._tab_layer_by_idx)))
            for child in dialog.findChildren(description._PseudoUSDBrowser):
                child._resolved_layers.clear()

            prim_index = parent.GetPrimIndex()
            _, sourcepath = tempfile.mkstemp()
            prim_index.DumpToDotGraph(sourcepath)
            targetpath = f"{sourcepath}.png"
            # create a temporary file loadable by our image tab
            _core_run([_core._which("dot"), sourcepath, "-Tpng", "-o", targetpath])
            browser._on_identifier_requested(anchor, targetpath)

            invalid_crate_layer = Sdf.Layer.CreateAnonymous()
            invalid_crate_layer.ImportFromString(
                # Not valid in USD-24.05: https://github.com/PixarAnimationStudios/OpenUSD/blob/59992d2178afcebd89273759f2bddfe730e59aa8/pxr/usd/sdf/testenv/testSdfParsing.testenv/baseline/127_varyingRelationship.sdf#L9
                """#sdf 1.4.32
                def GprimSphere "Sphere"
                {
                    delete varying rel constraintTarget = </Pivot3>
                    add varying rel constraintTarget = [
                        </Pivot3>,
                        </Pivot2>,
                    ]
                    reorder varying rel constraintTarget = [
                        </Pivot2>,
                        </Pivot>,
                    ]
                    varying rel constraintTarget.default = </Pivot>    
                }
                """
            )
            description._start_content_browser([invalid_crate_layer], None, stage.GetPathResolverContext(), ())

        with mock.patch("grill.views.description._which") as patch:
            patch.return_value = None
            with self.assertRaisesRegex(ValueError, "Expected arguments to contain an executable value on the first index"):
                description._start_content_browser(*args)

        error, result = _core._run(["python", 42])
        self.assertTrue(error.startswith('expected str'))
        self.assertEqual(result, "")

    def test_display_color_editor(self):
        stage = cook.fetch_stage(self.rootf)
        sphere = UsdGeom.Sphere.Define(stage, "/volume")
        color_var = sphere.GetDisplayColorPrimvar()
        editor = _attributes._DisplayColorEditor(color_var)
        editor._update_value()

        color_var.SetInterpolation(UsdGeom.Tokens.vertex)
        editor = _attributes._DisplayColorEditor(color_var)
        editor._update_value()
        editor._random.click()

        xform = stage.DefinePrim("/x")
        primvar = UsdGeom.Gprim(xform.GetPrim()).CreateDisplayColorPrimvar()
        editor = _attributes._DisplayColorEditor(primvar)
        with self.assertRaises(TypeError):  # atm some gprim types are not supported
            editor._update_value()

    def test_stats(self):
        empty = stats.StageStats()
        self.assertEqual(empty._usd_tree.topLevelItemCount(), 0)

        widget = stats.StageStats(stage=self.world)
        self.assertGreater(widget._usd_tree.topLevelItemCount(), 1)
        current = _qt.QtCharts
        del _qt.QtCharts
        stats.StageStats(stage=self.world)
        _qt.QtCharts = current


class TestGraphicsViewport(unittest.TestCase):
    def setUp(self):
        self._app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])

    def tearDown(self):
        self._app.quit()

    def test_zoom(self):
        """Zoom is triggered by ctrl + mouse wheel"""
        view = _graph._GraphicsViewport()

        initial_scale = view.transform().m11()

        position = QtCore.QPoint(10, 10)
        pixelDelta = QtCore.QPoint(0, 0)
        angleDelta_zoomIn = QtCore.QPoint(0, 120)
        buttons = QtCore.Qt.NoButton
        modifiers = QtCore.Qt.ControlModifier
        phase = QtCore.Qt.NoScrollPhase
        inverted = False

        # ZOOM IN
        event = QtGui.QWheelEvent(position, position, pixelDelta, angleDelta_zoomIn, buttons, modifiers, phase, inverted)
        view.wheelEvent(event)

        zoomed_in_scale = view.transform().m11()

        # Assert that the scale has changed according to the zoom logic
        self.assertGreater(zoomed_in_scale, initial_scale)
        angleDelta_zoomOut = QtCore.QPoint(-120, 0)

        # ZOOM OUT
        event = QtGui.QWheelEvent(position, position, pixelDelta, angleDelta_zoomOut, buttons, modifiers, phase, inverted)
        view.wheelEvent(event)
        self.assertGreater(zoomed_in_scale, view.transform().m11())

    def test_horizontal_scroll(self):
        """Horizontal scrolling with alt + mouse wheel"""
        view = _graph._GraphicsViewport()
        scroll_bar = view.horizontalScrollBar()
        initial_value = scroll_bar.value()
        scroll_bar.setMaximum(200)
        position = QtCore.QPoint(10, 10)
        pixelDelta = QtCore.QPoint(0, 0)
        angleDelta= QtCore.QPoint(-120, 0)
        buttons = QtCore.Qt.NoButton
        modifiers = QtCore.Qt.AltModifier
        phase = QtCore.Qt.NoScrollPhase
        inverted = False
        event = QtGui.QWheelEvent(position, position, pixelDelta, angleDelta, buttons, modifiers, phase, inverted)
        view.wheelEvent(event)
        final_value = scroll_bar.value()
        # Assert that the horizontal scroll has changed according to your pan logic
        self.assertGreater(final_value, initial_value)

    def test_vertical_scroll(self):
        """Vertical scroll with only mouse wheel"""
        view = _graph._GraphicsViewport()
        scroll_bar = view.verticalScrollBar()
        initial_value = scroll_bar.value()
        scroll_bar.setMaximum(200)
        position = QtCore.QPoint(10, 10)
        pixelDelta = QtCore.QPoint(0, 0)
        angleDelta = QtCore.QPoint(0, -120)
        buttons = QtCore.Qt.NoButton
        modifiers = QtCore.Qt.NoModifier
        phase = QtCore.Qt.NoScrollPhase
        inverted = False
        event = QtGui.QWheelEvent(position, position, pixelDelta, angleDelta, buttons, modifiers, phase, inverted)
        view.wheelEvent(event)
        final_value = scroll_bar.value()
        # Assert that the horizontal scroll has changed according to your pan logic
        self.assertGreater(final_value, initial_value)

    def test_pan(self):
        """Horizontal and vertical pan with mouse middle button"""
        view = _graph._GraphicsViewport()
        vertical_scroll_bar = view.verticalScrollBar()
        vertical_scroll_bar.setMaximum(200)
        horizontal_scroll_bar = view.horizontalScrollBar()
        horizontal_scroll_bar.setMaximum(200)
        start_position = QtCore.QPoint(50, 50)
        end_position = QtCore.QPoint(-5, -5)

        # 1. Mouse press
        middle_button_event = QtGui.QMouseEvent(QtCore.QEvent.MouseButtonPress, start_position, QtCore.Qt.MiddleButton, QtCore.Qt.MiddleButton, QtCore.Qt.NoModifier)
        vertical_value = vertical_scroll_bar.value()
        horizontal_value = horizontal_scroll_bar.value()
        view.mousePressEvent(middle_button_event)
        self.assertEqual(self._app.overrideCursor().shape(), QtGui.Qt.ClosedHandCursor)
        self.assertTrue(view._dragging)

        # 2. Mouse move
        view._last_pan_pos = _graph._EVENT_POSITION_FUNC(middle_button_event) + QtCore.QPoint(10,10)
        move_event = QtGui.QMouseEvent(QtCore.QEvent.MouseMove, end_position, QtCore.Qt.MiddleButton, QtCore.Qt.MiddleButton, QtCore.Qt.NoModifier)
        view.mouseMoveEvent(move_event)
        last_vertical_scroll_bar = vertical_scroll_bar.value()
        last_horizontal_scroll_bar = horizontal_scroll_bar.value()
        self.assertGreater(last_vertical_scroll_bar, vertical_value)
        self.assertGreater(last_horizontal_scroll_bar, horizontal_value)

        # 3. Release
        view.mouseReleaseEvent(middle_button_event)
        view._last_pan_pos = _graph._EVENT_POSITION_FUNC(middle_button_event) + QtCore.QPoint(20, 20)
        view.mouseMoveEvent(move_event)
        # Confirm no further move is performed
        self.assertEqual(last_vertical_scroll_bar, vertical_scroll_bar.value())
        self.assertEqual(last_horizontal_scroll_bar, horizontal_scroll_bar.value())



import uuid
import logging
import unittest
import tempfile

from pathlib import Path

from pxr import Usd, UsdGeom, Sdf, Ar, UsdUtils

from grill import cook, names, usd as gusd, tokens

logger = logging.getLogger(__name__)

# 2024-02-03 - Python-3.12 & USD-23.11
# python -m unittest --durations 0 test_cook
# Slowest test durations
# ----------------------------------------------------------------------
# 0.058s     test_define_taxon (test_cook.TestCook.test_define_taxon)
# 0.056s     test_inherited_and_specialized_contexts (test_cook.TestCook.test_inherited_and_specialized_contexts)
# 0.050s     test_create_on_previous_stage (test_cook.TestCook.test_create_on_previous_stage)
# 0.047s     test_asset_unit (test_cook.TestCook.test_asset_unit)
# 0.034s     test_spawn_unit (test_cook.TestCook.test_spawn_unit)
# 0.034s     test_spawn_unit_with_absolute_paths (test_cook.TestCook.test_spawn_unit_with_absolute_paths)
# 0.033s     test_create_many (test_cook.TestCook.test_create_many)
# 0.032s     test_spawn_many (test_cook.TestCook.test_spawn_many)
# 0.023s     test_fetch_stage (test_cook.TestCook.test_fetch_stage)
# 0.007s     test_edit_context (test_cook.TestCook.test_edit_context)
# 0.006s     test_match (test_cook.TestCook.test_match)
# 0.001s     test_spawn_many_invalid (test_cook.TestCook.test_spawn_many_invalid)
#
# ----------------------------------------------------------------------
# Ran 12 tests in 0.385s


class TestCook(unittest.TestCase):
    def setUp(self) -> None:
        tempdir = tempfile.mkdtemp()
        logger.debug(f"Repository root directory: {tempdir}")
        self.root_asset = names.UsdAsset.get_anonymous()
        self.token = cook.Repository.set(Path(tempdir) / "repo")

    def tearDown(self) -> None:
        cook.Repository.reset(self.token)

    def test_fetch_stage(self):
        root_asset = self.root_asset

        # TODO: cleanup this test. fetch_stage used to keep stages in a cache but not anymore.
        root_stage = cook.fetch_stage(root_asset)
        # fetching stage outside of AR _context should resolve to same stage
        self.assertEqual(root_stage.GetRootLayer().identifier, root_asset.name)

        repo_path = cook.Repository.get()
        resolver_ctx = Ar.DefaultResolverContext([str(repo_path)])
        with Ar.ResolverContextBinder(resolver_ctx):
            # inside an AR resolver _context, a new layer and custom stage should end up
            # in that stage not resolving to the same as the one from write.fetch_stage
            usd_opened = str(names.UsdAsset.get_anonymous(item='usd_opened'))
            Sdf.Layer.CreateNew(str(repo_path / usd_opened))
            non_cache_stage = Usd.Stage.Open(usd_opened)
            cached_stage = cook.fetch_stage(usd_opened)
            self.assertIsNot(non_cache_stage, cached_stage)
            # Even after fetching once, subsequent fetches should be different
            self.assertIsNot(cached_stage, cook.fetch_stage(usd_opened))

            # creating a new layer + stage + adding it to the cache manually
            # should still have fetch_stage to retrieve a different stage.
            sdf_opened = str(names.UsdAsset.get_default(item='sdf_opened'))
            Sdf.Layer.CreateNew(str(repo_path / sdf_opened))
            cached_layer = Sdf.Layer.FindOrOpen(sdf_opened)
            opened_stage = Usd.Stage.Open(cached_layer)
            cache = UsdUtils.StageCache.Get()
            cache.Insert(opened_stage)
            self.assertIsNot(opened_stage, cook.fetch_stage(sdf_opened))

        in_memory = Usd.Stage.CreateInMemory()
        from_memory = str(names.UsdAsset.get_anonymous(item='from_memory'))
        with self.assertRaises(ValueError):
            # a stage with an empty resolver that fetches a valid identifier should fail.
            cook.fetch_stage(from_memory, context=in_memory.GetPathResolverContext())

        unbound_resolver = str(names.UsdAsset.get_anonymous(item='unbound_resolver'))
        with self.assertRaises(RuntimeError):
            # directly fetching a new layer without a context with statement should fail
            cook._fetch_layer(unbound_resolver, root_stage.GetPathResolverContext())

    def test_match(self):
        root_stage = cook.fetch_stage(self.root_asset)
        with self.assertRaises(ValueError):
            cook._find_layer_matching(dict(missing='tokens'), root_stage.GetLayerStack())

    def test_edit_context(self):
        with self.assertRaises(TypeError):
            gusd.edit_context(object(), cook.fetch_stage(self.root_asset))

    def test_define_taxon(self):
        # An anonymous stage (non grill anonymous) should fail to define taxon.
        anon_stage = Usd.Stage.CreateInMemory()
        with self.assertRaises(ValueError):
            cook.define_taxon(anon_stage, "ShouldFail")
        # Same stage containing a grill anon layer on its stack should succeed.
        anon_pipeline = cook.fetch_stage(names.UsdAsset.get_anonymous())
        anon_stage.GetRootLayer().subLayerPaths.append(anon_pipeline.GetRootLayer().realPath)
        self.assertTrue(cook.define_taxon(anon_stage, "ShouldSucceed").IsValid())

        # Now, test stages fetched from the start via "common" pipeline calls.
        root_stage = cook.fetch_stage(self.root_asset)

        with self.assertRaisesRegex(ValueError, "reserved name"):
            cook.define_taxon(root_stage, cook._TAXONOMY_NAME)

        with self.assertRaisesRegex(ValueError, "reserved id fields"):
            cook.define_taxon(root_stage, "taxonomy_not_allowed", id_fields={cook._TAXONOMY_UNIQUE_ID: "by_id_value"})

        with self.assertRaisesRegex(ValueError, "reserved id fields"):
            cook.define_taxon(root_stage, "taxonomy_not_allowed", id_fields={cook._TAXONOMY_UNIQUE_ID.name: "by_id_name"})

        with self.assertRaisesRegex(ValueError, "invalid id_field keys"):
            cook.define_taxon(root_stage, "nonexistingfield", id_fields={str(uuid.uuid4()): "by_id_name"})

        displayable = cook.define_taxon(root_stage, "DisplayableName")
        # idempotent call should keep previously created prim
        self.assertEqual(displayable, cook.define_taxon(root_stage, "DisplayableName"))

        person = cook.define_taxon(root_stage, "Person", references=(displayable,))

        with cook.taxonomy_context(root_stage):
            displayable.CreateAttribute("label", Sdf.ValueTypeNames.String)

        missing_or_empty_fields_msg = f"Missing or empty '{cook._FIELDS_KEY}'"
        not_taxon = root_stage.DefinePrim("/not/a/taxon")
        with self.assertRaisesRegex(ValueError, missing_or_empty_fields_msg):
            cook.create_unit(not_taxon, "WillFail")

        not_taxon.SetAssetInfoByKey(cook._ASSETINFO_KEY, {})
        with self.assertRaisesRegex(ValueError, missing_or_empty_fields_msg):
            cook.create_unit(not_taxon, "WillFail")

        not_taxon.SetAssetInfoByKey(cook._ASSETINFO_KEY, {'invalid': 42})
        with self.assertRaisesRegex(ValueError, missing_or_empty_fields_msg):
            cook.create_unit(not_taxon, "WillFail")

        not_taxon.SetAssetInfoByKey(cook._ASSETINFO_KEY, {cook._FIELDS_KEY: 42})
        with self.assertRaisesRegex(TypeError, f"Expected mapping on key '{cook._FIELDS_KEY}'"):
            cook.create_unit(not_taxon, "WillFail")

        not_taxon.SetAssetInfoByKey(cook._ASSETINFO_KEY, {cook._FIELDS_KEY: {}})
        with self.assertRaisesRegex(ValueError, missing_or_empty_fields_msg):
            cook.create_unit(not_taxon, "WillFail")

        emil = cook.create_unit(person, "EmilSinclair", label="Emil Sinclair")
        self.assertEqual(emil, cook.create_unit(person, "EmilSinclair"))

        with cook.unit_context(emil):
            emil.GetVariantSet("Transport").SetVariantSelection("HorseDrawnCarriage")

        hero = cook.define_taxon(root_stage, "Hero", references=(person,))
        batman = cook.create_unit(hero, "Batman")
        expected_people = [emil, batman]  # batman is also a person
        expected_heroes = [batman]
        stage_prims = root_stage.Traverse()
        self.assertEqual(expected_people, list(cook.itaxa(stage_prims, person)))
        self.assertEqual(expected_heroes, list(cook.itaxa(stage_prims, hero)))

    def test_create_on_previous_stage(self):
        """Confirm that creating assets on a previously saved stage works.

        The default behavior from layer identifiers that are relative to the resolver search path is to be absolute
        when a stage using them is re-opened, so:
            original_identifier.usda
                becomes
            /absolute/path/original_identifier.usda
        """
        root_asset = names.UsdAsset.get_anonymous()
        root_stage = cook.fetch_stage(root_asset)
        # creates taxonomy.usda and adds it to the stage layer stack
        cook.define_taxon(root_stage, "FirstTaxon")
        root_stage.Save()
        del root_stage

        reopened_stage = cook.fetch_stage(root_asset)
        # the taxonomy.usda now has as identifier /absolute/path/taxonomy.usda, so confirm we can use it still
        cook.create_many(cook.define_taxon(reopened_stage, "SecondTaxon"), ["A", "B"])

    def test_asset_unit(self):
        stage = cook.fetch_stage(self.root_asset)
        taxon_name = "Person"
        person = cook.define_taxon(stage, taxon_name)
        unit_name = "EmilSinclair"
        emil = cook.create_unit(person, unit_name, label="Emil Sinclair")
        unit_asset = cook.unit_asset(emil)
        unit_id = names.UsdAsset(unit_asset.identifier)
        self.assertEqual(unit_name, getattr(unit_id, cook._UNIT_UNIQUE_ID.name))
        self.assertEqual(taxon_name, getattr(unit_id, cook._TAXONOMY_UNIQUE_ID.name))

        not_a_unit = stage.DefinePrim(emil.GetPath().AppendChild("not_a_unit"))
        with self.assertRaisesRegex(ValueError, "Missing or empty"):
            cook.unit_asset(not_a_unit)

        layer = Sdf.Layer.CreateAnonymous()
        with self.assertRaisesRegex(ValueError, "Could not find appropriate node for edit target"):
            gusd.edit_context(not_a_unit, Usd.PrimCompositionQuery.Filter(), lambda arc: arc.GetTargetNode().layerStack.identifier.rootLayer == layer)

        # break the unit model API
        Usd.ModelAPI(emil).SetAssetIdentifier("")
        without_modelapi = cook.unit_asset(emil)
        self.assertEqual(unit_asset, without_modelapi)  # we should get the same result

        Usd.ModelAPI(emil).SetAssetName("not_emil")
        with self.assertRaisesRegex(ValueError, "Could not find layer matching"):
            cook.unit_asset(emil)

    def test_create_many(self):
        stage = cook.fetch_stage(self.root_asset)
        taxon = cook.define_taxon(stage, "Anon")
        cook.create_many(taxon, ("first", "second"))

    def test_spawn_unit(self):
        stage = cook.fetch_stage(self.root_asset)
        id_fields = {tokens.ids.CGAsset.kingdom.name: "K"}
        taxon = cook.define_taxon(stage, "Another", id_fields=id_fields)
        parent, child = cook.create_many(taxon, ['A', 'B'])
        with cook.unit_context(parent):
            for path, value in (
                    ("", (2, 15, 6)),
                    ("Deeper/Nested/Golden1", (-4, 5, 1)),
                    ("Deeper/Nested/Golden2", (-4, -10, 1)),
                    ("Deeper/Nested/Golden3", (0, 10, -2)),
            ):
                cook.spawn_unit(parent, child, path)

    def test_spawn_unit_with_absolute_paths(self):
        stage = cook.fetch_stage(self.root_asset)
        id_fields = {tokens.ids.CGAsset.kingdom.name: "K"}
        taxon = cook.define_taxon(stage, "Another", id_fields=id_fields)
        parent, child = cook.create_many(taxon, ['A', 'B'])
        valid_path = parent.GetPath().AppendPath("Deeper/Nested/Golden1")
        invalid_path = "/invalid/path"
        self.assertTrue(cook.spawn_unit(parent, child, valid_path))
        with self.assertRaisesRegex(ValueError, "needs to be a child path of parent path"):
            cook.spawn_unit(parent, child, invalid_path)

    def test_spawn_many(self):
        stage = cook.fetch_stage(self.root_asset)
        id_fields = {tokens.ids.CGAsset.kingdom.name: "K"}
        taxon = cook.define_taxon(stage, "Another", id_fields=id_fields)
        parent, child = cook.create_many(taxon, ['A', 'B'])
        cook.spawn_many(parent, child, ["b"], labels=["1", "2"])
        self.assertEqual(len(parent.GetChildren()), 1)

    def test_spawn_many_invalid(self):
        stage = Usd.Stage.CreateInMemory()
        parent = stage.DefinePrim("/a")
        with self.assertRaisesRegex(ValueError, "Can not spawn .* to itself."):
            cook.spawn_many(parent, parent, ["impossible"])
        child = stage.DefinePrim("/b")  # child needs to be a grill unit
        with self.assertRaisesRegex(ValueError, "Could not extract identifier from"):
            cook.spawn_many(parent, child, ["b"])

    def test_inherited_and_specialized_contexts(self):
        stage = cook.fetch_stage(self.root_asset)
        id_fields = {tokens.ids.CGAsset.kingdom.name: "K"}
        taxon = cook.define_taxon(stage, "Another", id_fields=id_fields)
        parent, via_s, via_i, not_under_context = cook.create_many(taxon, ['parent', 'via_s', 'via_i', 'not_under_context'])

        not_a_unit = stage.DefinePrim("/vanilla_prim")
        with self.assertRaisesRegex(ValueError, "is not a valid unit"):
            cook.specialized_context(not_a_unit)

        with self.assertRaisesRegex(ValueError, "needs to be a valid unit"):
            cook.specialized_context(via_s, via_s.GetParent())

        with self.assertRaisesRegex(ValueError, "is not a descendant"):
            cook.specialized_context(parent, via_s)

        spawned_invalid = cook.spawn_unit(parent, not_under_context)
        with self.assertRaisesRegex(ValueError, "Is there a composition arc bringing"):
            # TODO: find a more meaningful message (higher level) than the edit target context one.
            cook.specialized_context(spawned_invalid, parent)

        with cook.unit_context(parent):
            via_s_spawned = cook.spawn_unit(parent, via_s)
            via_i_spawned = cook.spawn_unit(parent, via_i)

        with cook.inherited_context(not_under_context):
            UsdGeom.Gprim(not_under_context).MakeInvisible()

        with cook.specialized_context(via_s_spawned, parent):
            UsdGeom.Gprim(via_s_spawned).MakeInvisible()

        with cook.inherited_context(via_i_spawned):
            UsdGeom.Gprim(via_i_spawned).MakeInvisible()

        def _check_broadcasted_invisibility(asset, prim, method):
            target_stage = Usd.Stage.Open(asset)
            target_prefix = cook._broadcast_root_path(prim, method)
            authored = UsdGeom.Gprim(target_stage.GetPrimAtPath(target_prefix.AppendChild(Usd.ModelAPI(prim).GetAssetName()))).GetVisibilityAttr().Get()
            self.assertEqual(authored, 'invisible')

        for target_asset, target_prim, broadcast_type in (
                (not_under_context, not_under_context, Usd.Inherits),  # non-referenced, no context, asset unit is the target
                (via_i_spawned, via_i_spawned, Usd.Inherits),  # referenced asset unit, no context, asset unit is the target
                (parent, via_s_spawned, Usd.Specializes),  # referenced, context unit is the target
        ):
            with self.subTest(target_asset=str(target_asset), target_prim=str(target_prim), broadcast_type=str(broadcast_type)):
                _check_broadcasted_invisibility(cook.unit_asset(target_asset), target_prim, broadcast_type)



import unittest

from pxr import Usd, UsdGeom, Sdf

import grill.usd as gusd


class TestUSD(unittest.TestCase):
    def test_edit_context(self):
        main = Usd.Stage.CreateInMemory()
        # Jump between 3 different layer stacks adding variants to the same set
        # main [variant blue] -> reference [variant green] -> payload [variant red]
        referenced = Usd.Stage.CreateInMemory()
        referenced.SetDefaultPrim(referenced.DefinePrim("/Referenced"))
        reference = Sdf.Reference(referenced.GetRootLayer().identifier)

        payloaded = Usd.Stage.CreateInMemory()
        payloaded.SetDefaultPrim(payloaded.DefinePrim("/Payloaded"))
        payload = Sdf.Payload(payloaded.GetRootLayer().identifier)

        top = main.DefinePrim("/Top")
        main.SetDefaultPrim(top)
        top.GetReferences().AddReference(reference)

        with gusd.edit_context(reference, top):
            top.GetPayloads().AddPayload(payload)
            with gusd.edit_context(payload, top):
                geom = UsdGeom.Sphere.Define(main, top.GetPath().AppendPath("inner/child"))
                color = geom.GetDisplayColorAttr()
                color_set = geom.GetPrim().GetVariantSets().AddVariantSet("color")
                color_set.AddVariant("from_payload")
                color_set.SetVariantSelection("from_payload")
                with gusd.edit_context(color_set, payloaded.GetRootLayer()):  # color_set.GetVariantEditContext() would fail here
                    color.Set([(1,0,0)])
                color_set.ClearVariantSelection()
            color_set.AddVariant("from_reference")
            color_set.SetVariantSelection("from_reference")
            with gusd.edit_context(color_set, referenced.GetRootLayer()):
                color.Set([(0,1,0)])
            color_set.ClearVariantSelection()

        color_set.AddVariant("from_top")
        color_set.SetVariantSelection("from_top")
        with color_set.GetVariantEditContext():
            color.Set([(0,0,1)])

        color_set.ClearVariantSelection()

        self.assertIn(payload, referenced.GetRootLayer().GetPrimAtPath(referenced.GetDefaultPrim().GetPath()).payloadList.GetAddedOrExplicitItems())
        self.assertIn(reference, main.GetRootLayer().GetPrimAtPath(top.GetPath()).referenceList.GetAddedOrExplicitItems())

        for stage, variant_name in (
            (main, "from_top"),
            (referenced, "from_reference"),
            (payloaded, "from_payload")
        ):
            layer = stage.GetRootLayer()
            self.assertIsNotNone(layer.GetPrimAtPath(f"{layer.defaultPrim}/inner/child{{color={variant_name}}}"))

    def test_missing_arc(self):
        stage = Usd.Stage.CreateInMemory()
        prim = stage.DefinePrim("/Referenced")
        anon = Usd.Stage.CreateInMemory()
        reference = Sdf.Reference(anon.GetRootLayer().identifier)
        with self.assertRaisesRegex(ValueError, "without a prim path"):
            gusd.edit_context(reference, prim)

    def test_missing_layer(self):
        stage = Usd.Stage.CreateInMemory()
        prim = stage.DefinePrim("/Referenced")
        reference = Sdf.Reference("non_existing")
        with self.assertRaisesRegex(ValueError, "ability to find layer"):
            gusd.edit_context(reference, prim)

    def test_format_tree(self):
        stage = Usd.Stage.CreateInMemory()
        child1 = stage.DefinePrim("/path/to/child1")
        stage.DefinePrim("/path/to/child02")
        self.assertEqual('┐to\n├── child1\n└── child02', gusd._format_prim_hierarchy([child1.GetParent()]))
        self.assertEqual('┐to', gusd._format_prim_hierarchy([child1.GetParent()], include_descendants=False))
        self.assertEqual('┐/\n└── path\n    └── to\n        ├── child1\n        └── child02', gusd._format_prim_hierarchy([stage.GetPseudoRoot()]))
        self.assertEqual('┐/', gusd._format_prim_hierarchy([stage.GetPseudoRoot()], include_descendants=False))

    def test_make_plane(self):
        """Not a public API yet"""
        stage = Usd.Stage.CreateInMemory()
        width, depth = 10, 8
        mesh = UsdGeom.Mesh.Define(stage, "/m")
        gusd._make_plane(mesh, width, depth)
        self.assertEqual(80, len(mesh.GetPointsAttr().Get()))
        self.assertEqual(252, len(mesh.GetFaceVertexIndicesAttr().Get()))
        self.assertEqual(63, len(mesh.GetFaceVertexCountsAttr().Get()))



"""Shared members for views modules, not considered public API."""
import os
import enum
import shutil
import typing
import subprocess
import contextlib
from pathlib import Path
from functools import partial, cache

from ._qt import QtWidgets, QtGui, QtCore

# Agreement: raw data accessible here
_QT_OBJECT_DATA_ROLE = QtCore.Qt.UserRole + 1

# Very slightly modified USDView stylesheet for the push buttons.
_USDVIEW_PUSH_BUTTON_STYLE = """
QPushButton{
    /* gradient background */
    background-color: QLinearGradient(x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 rgb(100, 100, 100), stop: 1 rgb(90, 90, 90));

    /* thin dark round border */
    border-width: 1px;
    border-color: rgb(42, 42, 42);
    border-style: solid;
    border-radius: 3;

    /* give the text enough space */
    padding: 3px;
    padding-right: 10px;
    padding-left: 10px;
}

/* Darker gradient when the button is pressed down */
QPushButton:pressed {
    background-color: QLinearGradient(x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 rgb(50, 50, 50), stop: 1 rgb(60, 60, 60));
}

QPushButton:checked {
    background-color: QLinearGradient(x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 rgb(60, 65, 70), stop: 1 rgb(70, 75, 80));
}

/* Greyed-out colors when the button is disabled */
QPushButton:disabled {
    background-color: QLinearGradient(x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 rgb(66, 66, 66), stop: 1 rgb(56, 56, 56));
}

QCheckBox::indicator:indeterminate:hover {
    border: 1px solid rgb(163, 135, 78);
}
/* Partially checked state (indeterminate) */
QCheckBox::indicator:indeterminate {
    background-color: rgb(135, 206, 250); /* lightskyblue */
}

"""

# Taken from QTreeWidget style and adapted for _Tree:
# https://github.com/PixarAnimationStudios/USD/blob/3abc46452b1271df7650e9948fef9f0ce602e3b2/pxr/usdImaging/usdviewq/usdviewstyle.qss#L258
_USDVIEW_QTREEVIEW_STYLE = """
_Tree {
    alternate-background-color: rgb(59, 59, 59);
}

._Tree::item, QTableView::item {
    /* this border serves to separate the columns
     * since the grid is often invised */
    border-right: 1px solid rgb(41, 41, 41);
    padding-top: 1px;
    padding-bottom: 1px;
}

/* Selected items highlighted in orange */
._Tree::item:selected,
_Tree::branch:selected,
QTableView::item:selected {
    background: rgb(189, 155, 84);
}

/* hover items a bit lighter */
._Tree::item:hover:!pressed:!selected,
_Tree::branch:hover:!pressed:!selected,
QTableView::item:hover:!pressed:!selected {
    background: rgb(70, 70, 70);
}

._Tree::item:hover:!pressed:selected,
_Tree::branch:hover:!pressed:selected,
QTableView::item:hover:!pressed:selected {
/*    background: rgb(132, 109, 59); */
    background: rgb(227, 186, 101);
}
"""

_USDVIEW_STYLE = _USDVIEW_PUSH_BUTTON_STYLE + _USDVIEW_QTREEVIEW_STYLE

@cache
def _which(what):
    return shutil.which(what)


def _run(args: list):
    if not args or not args[0]:
        raise ValueError(f"Expected arguments to contain an executable value on the first index. Got: {args}")
    kwargs = dict(capture_output=True)
    if hasattr(subprocess, 'CREATE_NO_WINDOW'):  # Only on Windows OS
        kwargs.update(creationflags=subprocess.CREATE_NO_WINDOW)
    try:
        result = subprocess.run(args, **kwargs)
    except TypeError as exc:
        return str(exc), ""
    else:
        error = result.stderr.decode() if result.returncode else None
        return error, result.stdout.decode()


@cache
def _ensure_dot():
    """For usage only when DCC python interpreter fails to install pygraphviz properly."""
    if dotpath := _which("dot"):
        # https://github.com/pygraphviz/pygraphviz/issues/360
        # TODO: is this the best approach to solve current Houdini and Maya failing to import graphviz??
        if hasattr(os, "add_dll_directory"):
            os.add_dll_directory(Path(dotpath).parent)  # sigh, patch pygraphviz?


@contextlib.contextmanager
def wait():
    try:
        QtWidgets.QApplication.setOverrideCursor(QtGui.QCursor(QtCore.Qt.WaitCursor))
        yield
    finally:
        QtWidgets.QApplication.restoreOverrideCursor()


class _EMOJI(enum.Enum):  # Replace with StrEnum in 3.11
    # All emojis have an additional space at the end since Maya-2023.2 and Houdini-19.5 are unable to display emoji otherwise
    # GENERAL
    ID = "🕵 "
    VISIBILITY = "👀 "
    SEARCH = "🔎 "
    LOCK = "🔐 "
    UNLOCK = "🔓 "

    # STAGE TRAVERSAL
    MODEL_HIERARCHY = "🏡 "
    INSTANCE_PROXIES = "💠 "

    # PRIM SPECIFIER
    ORPHANED = "👻 "
    CLASSES = "🧪 "
    DEFINED = "🧱 "

    # PRIM STATUS
    ACTIVE = "💡 "
    INACTIVE = "🌒 "

    # IDENTIFICATION
    NAME = "🔖 "


class _Column(typing.NamedTuple):
    """Foundational structure for a column and how to retrieve / set model data."""
    name: str
    getter: callable = None
    setter: callable = None
    editor: callable = None


class _ColumnOptions(enum.Flag):
    """Options that will be available on the header columns of a view."""
    NONE = enum.auto()
    SEARCH = enum.auto()
    VISIBILITY = enum.auto()
    LOCK = enum.auto()
    ALL = SEARCH | VISIBILITY | LOCK


class _ColumnItemDelegate(QtWidgets.QStyledItemDelegate):
    """
    https://doc.qt.io/qtforpython/overviews/sql-presenting.html
    https://doc.qt.io/qtforpython/overviews/qtwidgets-itemviews-spinboxdelegate-example.html
    """

    def createEditor(self, parent: QtWidgets.QWidget, option: QtWidgets.QStyleOptionViewItem, index: QtCore.QModelIndex) -> QtWidgets.QWidget:
        creator = self.parent()._model._columns[index.column()].editor or super().createEditor
        return creator(parent, option, index)


class _EmptyItemDelegate(_ColumnItemDelegate):
    def setModelData(self, editor: QtWidgets.QWidget, model: QtCore.QAbstractItemModel, index: QtCore.QModelIndex):
        setter = self.parent()._model._columns[index.column()].setter or super().setModelData
        return setter(editor, model, index)


class _ColumnHeaderOptions(QtWidgets.QWidget):
    """A widget to be used within a header for columns on a table / tree view."""
    def __init__(self, name, options: _ColumnOptions, *args, **kwargs):
        super().__init__(*args, **kwargs)
        layout = QtWidgets.QVBoxLayout()
        options_layout = QtWidgets.QHBoxLayout()
        self.setLayout(layout)

        # Search filter
        self._line_filter = line_filter = QtWidgets.QLineEdit()
        line_filter.setPlaceholderText("Filter")
        # TODO: add functionality for "inverse regex"
        line_filter.setToolTip(r"Negative lookahead: ^((?!{expression}).)*$")

        # Visibility
        self._vis_button = vis_button = QtWidgets.QPushButton(_EMOJI.VISIBILITY.value)
        vis_button.setCheckable(True)
        vis_button.setChecked(True)
        vis_button.setFlat(True)

        # Lock
        self._lock_button = lock_button = QtWidgets.QPushButton()
        lock_button.setCheckable(True)
        lock_button.setFlat(True)

        # allow for a bit of extra space with option buttons
        label = QtWidgets.QLabel(f"{name} ")
        options_layout.addWidget(label)
        options_layout.addWidget(vis_button)
        options_layout.addWidget(lock_button)
        options_layout.addStretch()
        layout.addLayout(options_layout)
        self._filter_layout = filter_layout = QtWidgets.QFormLayout()
        filter_layout.addRow(_EMOJI.SEARCH.value, line_filter)
        if _ColumnOptions.SEARCH in options:
            layout.addLayout(filter_layout)
        self._options = options
        self._decorateLockButton(lock_button, lock_button.isChecked())
        lock_button.toggled.connect(partial(self._decorateLockButton, lock_button))

        # set visibility after widgets are added to our layout
        vis_button.setVisible(_ColumnOptions.VISIBILITY in options)
        lock_button.setVisible(_ColumnOptions.LOCK in options)

        # public members exposure
        self.label = label
        self.locked = self._lock_button.toggled
        self.toggled = self._vis_button.toggled
        self.filterChanged = line_filter.textChanged

    def resizeEvent(self, event:QtGui.QResizeEvent):
        """Update the widget mask after resize to bypass clicks to the parent widget."""
        value = super().resizeEvent(event)
        self._updateMask()
        return value

    def _updateMask(self):
        """We want nothing but the filter and buttons to be clickable on this widget."""
        region = QtGui.QRegion(self.frameGeometry())
        if _ColumnOptions.SEARCH in self._options:
            region += QtGui.QRegion(self._filter_layout.geometry())
        # when buttons are flat, geometry has a small render offset on x
        if _ColumnOptions.LOCK in self._options:
            region += self._lock_button.geometry().adjusted(-2, 0, 2, 0)
        if _ColumnOptions.VISIBILITY in self._options:
            region += self._vis_button.geometry().adjusted(-2, 0, 2, 0)
        self.setMask(region)

    def _decorateLockButton(self, button, locked):
        if locked:
            text = _EMOJI.LOCK.value
            tip = "Column is non-editable (locked).\nClick to allow edits."
        else:
            text = _EMOJI.UNLOCK.value
            tip = "Edits are allowed on this column (unlocked).\nClick to block edits."
        button.setText(text)
        button.setToolTip(tip)

    def _setHidden(self, value):
        self._vis_button.setChecked(not value)


class _Header(QtWidgets.QHeaderView):
    """A header that allows to display the column options for a table / tree.

    See also:
        https://www.qt.io/blog/2014/04/11/qt-weekly-5-widgets-on-a-qheaderview
        https://www.qt.io/blog/2012/09/28/qt-support-weekly-27-widgets-on-a-header
        https://www.learnpyqt.com/courses/model-views/qtableview-modelviews-numpy-pandas/
    """
    def __init__(self, columns: typing.Iterable[str], options: _ColumnOptions, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.options_by_index = dict()  # {int: _ColumnHeaderOptions}
        self._proxy_labels = dict()  # I've found no other way around this
        for index, name in enumerate(columns):
            column_options = _ColumnHeaderOptions(name, options=options, parent=self)
            column_options.layout().setContentsMargins(0, 0, 0, 0)
            self.options_by_index[index] = column_options
            self.setMinimumHeight(column_options.sizeHint().height() + 20)

            # we keep track of the column options label but our proxy will bypass clicks
            # allowing for UX when clicking on column headers
            proxy_label = QtWidgets.QLabel(parent=self)
            proxy_label.setText(column_options.label.text())
            proxy_label.setAttribute(QtCore.Qt.WA_TransparentForMouseEvents)
            self._proxy_labels[column_options.label] = proxy_label

        self.setStretchLastSection(True)
        self.setSectionsMovable(True)
        self.sectionResized.connect(self._handleSectionResized)
        self.sectionMoved.connect(self._handleSectionMoved)
        self.sectionClicked.connect(self._handleSectionClicked)

    def _updateOptionsGeometry(self, logical_index: int):
        """Updates the options geometry for the column at the logical index"""
        widget = self.options_by_index[logical_index]
        geometry = self._geometryForWidget(logical_index)
        widget.setGeometry(*geometry)
        label_geo = widget.label.geometry()
        label_geo.moveTo(widget.pos())
        self._proxy_labels[widget.label].setGeometry(label_geo)

    def _updateVisualSections(self, start_index):
        """Updates all of the sections starting at the given index."""
        for index in range(start_index, self.count()):
            if self.isSectionHidden(index):
                continue
            self._updateOptionsGeometry(self.logicalIndex(index))

    def showEvent(self, event:QtGui.QShowEvent):
        for index, widget in self.options_by_index.items():
            self._updateOptionsGeometry(index)
            # ensure we have readable columns upon show
            widget.show()
            self.resizeSection(index, widget.sizeHint().width() + 20)
        super().showEvent(event)

    def _handleSectionResized(self, index):
        self._updateVisualSections(self.visualIndex(index))
        for index, widget in self.options_by_index.items():
            # if new size is smaller than the width hint half, make options invisible
            vis = widget.minimumSizeHint().width() / 2.0 < self.sectionSize(index)
            widget.setVisible(vis)
            self._proxy_labels[widget.label].setVisible(vis)

    def _handleSectionMoved(self, __, old_visual_index, new_visual_index):
        self._updateVisualSections(min(old_visual_index, new_visual_index))

    def _handleSectionClicked(self, index):
        """Without this, when a section is clicked (e.g. when sorting),
        we'd have a mismatch on the proxy geometry label.
        """
        # With big stages (teste with ~51k prims) calling directly self._updateVisualSections does not adjust sizes.
        # I've only found a way to guarantee the update by delaying execution with a QTimer.
        QtCore.QTimer.singleShot(1, lambda: self._updateVisualSections(0))

    def _geometryForWidget(self, index):
        """Main geometry for the widget to show at the given index"""
        return self.sectionViewportPosition(index) + 10, 10, self.sectionSize(index) - 20, self.height() - 20


class EmptyTableModel(QtGui.QStandardItemModel):
    """Minimal empty table for new data (unlike existing USD stages or layers).

    This is a "transient" model which will eventually have data translated into USD.
    """
    def __init__(self, columns, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._columns = columns
        self._locked_columns = set()  # TODO: make sure this plays well with this and USD table
        self.setHorizontalHeaderLabels([''] * len(columns))


class _ObjectTableModel(QtCore.QAbstractTableModel):
    """Table model objects whose getters / setters are provided via columns.

    Mainly used for USD objects like layers or prims.
    """

    def __init__(self, columns, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._columns = columns
        self._locked_columns = set()
        self._objects = []

    def rowCount(self, parent:QtCore.QModelIndex=...) -> int:
        return len(self._objects)

    def columnCount(self, parent:QtCore.QModelIndex=...) -> int:
        return len(self._columns)

    def data(self, index:QtCore.QModelIndex, role:int=...) -> typing.Any:
        if role == _QT_OBJECT_DATA_ROLE:  # raw data
            return self._objects[index.row()]
        elif role == QtCore.Qt.DisplayRole:
            obj = self.data(index, role=_QT_OBJECT_DATA_ROLE)
            return self._columns[index.column()].getter(obj)
        elif role == QtCore.Qt.EditRole:
            obj = self.data(index, role=_QT_OBJECT_DATA_ROLE)
            return self._columns[index.column()].getter(obj)

    def sort(self, column:int, order:QtCore.Qt.SortOrder=...) -> None:
        self.layoutAboutToBeChanged.emit()
        key = self._columns[column].getter
        reverse = order == QtCore.Qt.SortOrder.AscendingOrder
        try:
            self._objects = sorted(self._objects, key=key, reverse=reverse)
        finally:
            self.layoutChanged.emit()

    def setData(self, index:QtCore.QModelIndex, value:typing.Any, role:int=...) -> bool:
        obj = self.data(index, role=_QT_OBJECT_DATA_ROLE)
        result = self._columns[index.column()].setter(obj, value)
        print(f"Result: {result}")
        # self.dataChanged.emit(topLeft, bottomRight)  # needed?
        return True


class _ProxyModel(QtCore.QSortFilterProxyModel):
    def headerData(self, section: int, orientation: QtCore.Qt.Orientation, role: int = ...):
        """For a vertical header, display a sequential visual index instead of the logical from the model."""
        # https://www.walletfox.com/course/qsortfilterproxymodelexample.php
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Vertical:
                return section + 1
            elif orientation == QtCore.Qt.Horizontal:
                return ""  # our horizontal header labels are drawn by custom header
        return super().headerData(section, orientation, role)

    def sort(self, column: int, order: QtCore.Qt.SortOrder = QtCore.Qt.AscendingOrder) -> None:
        self.sourceModel().sort(column, order)


class _ColumnHeaderMixin:
    # TODO: see if this makes sense.
    def __init__(self, model, columns: typing.Iterable[_Column], options: _ColumnOptions, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # TODO: I tried splitting this into a setModel method but keeping a self reference
        #   for columns or options causes python to segfault on Travis (not locally)
        # self._columns = columns
        # self._options = options
        self._model = model
        header = _Header([col.name for col in columns], options, QtCore.Qt.Horizontal)

        # TODO: item delegate per model type? For now it works ):<
        column_delegate_cls = _ColumnItemDelegate if isinstance(model, _ObjectTableModel) else _EmptyItemDelegate
        self._column_options = header.options_by_index

        # for every column, create a proxy model and chain it to the next one
        for column_index, column_data in enumerate(columns):
            proxy_model = _ProxyModel()
            proxy_model.setSourceModel(model)
            proxy_model.setFilterKeyColumn(column_index)

            column_options = header.options_by_index[column_index]
            if _ColumnOptions.SEARCH in options:
                self._connect_search(column_options, column_index, proxy_model)
            if _ColumnOptions.VISIBILITY in options:
                self._connect_visibility(column_options, column_index, proxy_model)
            if _ColumnOptions.LOCK in options:
                self._connect_locked(column_options, column_index, proxy_model)

            delegate = column_delegate_cls(parent=self)
            self.setItemDelegateForColumn(column_index, delegate)

            model = proxy_model

        header.setModel(model)
        header.setSectionsClickable(True)

        self.setModel(model)
        try:
            self.setHorizontalHeader(header)
        except AttributeError:
            self.setHeader(header)

    def _connect_search(self, options, index, model):
        options.filterChanged.connect(model.setFilterRegularExpression)

    def _connect_visibility(self, options, index, model):
        options.toggled.connect(partial(self._setColumnVisibility, index))

    def _connect_locked(self, options, index, model):
        options.locked.connect(partial(self._setColumnLocked, index))

    def _setColumnVisibility(self, index: int, visible: bool):
        self.setColumnHidden(index, not visible)
        self._fixPositions()

    def _setColumnLocked(self, column_index, value):
        method = set.add if value else set.discard
        method(self._model._locked_columns, column_index)

    def scrollContentsBy(self, dx:int, dy:int):
        super().scrollContentsBy(dx, dy)
        if dx:
            self._fixPositions()

    def _fixPositions(self):
        try:
            header = self.horizontalHeader()  # Tables
        except AttributeError:
            header = self.header()  # Trees
        header._updateVisualSections(min(header.options_by_index))



import io
import csv
import enum
import typing
import weakref
import inspect
import logging
import operator
import textwrap
import itertools
from functools import lru_cache
from collections import Counter

from pxr import Usd, UsdGeom, Sdf
from ._qt import QtCore, QtWidgets, QtGui

from .. import usd as _usd
from . import _core

logger = logging.getLogger(__name__)

# {Mesh: UsdGeom.Mesh, Xform: UsdGeom.Xform}
# TODO: add more types here
_PRIM_TYPE_OPTIONS = dict(x for x in inspect.getmembers(UsdGeom, inspect.isclass) if Usd.Typed in x[-1].mro())


def _prim_type_combobox(parent, option, index):
    # TODO: see if we can lru_cache as well
    combobox = QtWidgets.QComboBox(parent=parent)
    combobox.addItems(sorted(_PRIM_TYPE_OPTIONS))
    return combobox


@lru_cache(maxsize=None)
def _traverse_predicate(*, model_hierarchy: bool = False, instance_proxies: bool = False):
    predicate = Usd.PrimIsModel if model_hierarchy else Usd.PrimAllPrimsPredicate
    return Usd.TraverseInstanceProxies(predicate) if instance_proxies else predicate


@lru_cache(maxsize=None)
def _filter_predicate(*, orphaned, classes, defined, active, inactive, logical_op):
    def specifier(prim):
        is_class = prim.IsAbstract()
        is_defined = prim.IsDefined()
        return ((classes and is_class) or
                (orphaned and not is_defined) or
                (defined and is_defined and not is_class))

    def status(prim):
        is_active = prim.IsActive()
        return (active and is_active) or (inactive and not is_active)

    return lambda prim: logical_op(specifier(prim), status(prim))


@lru_cache(maxsize=None)
def _prim_font(*, abstract=False, orphaned=False):
    # Passing arguments via the constructor does not work even though docs say they should
    font = QtGui.QFont()
    # Abstract prims are also considered defined; since we want to distinguish abstract
    # defined prims from non-abstract defined prims, we check for abstract first.
    if abstract:
        font.setWeight(QtGui.QFont.ExtraLight)
        font.setLetterSpacing(QtGui.QFont.PercentageSpacing, 120)
    elif orphaned:
        font.setWeight(QtGui.QFont.Light)
        font.setItalic(True)
    else:
        font.setWeight(QtGui.QFont.Normal)
    return font


class _PrimTextColor(enum.Enum):
    NONE = None
    INSTANCE = QtGui.QColor('lightskyblue')
    PROTOTYPE = QtGui.QColor('cornflowerblue')
    INACTIVE = QtGui.QColor('darkgray')
    INACTIVE_ARCS = QtGui.QColor('orange').darker(150)
    ARCS = QtGui.QColor('orange')


class StageTableModel(_core._ObjectTableModel):
    """This model provides flexibility for:

    - Specifying traversal method
    - Specifying root prims to traverse
    - Specifying prim paths to prune from traversal
    - Filtering prims based on a provided predicate
    """
    # https://doc.qt.io/qtforpython/PySide6/QtCore/QAbstractItemModel.html
    # https://doc.qt.io/qtforpython/overviews/qtwidgets-itemviews-pixelator-example.html#pixelator-example
    # I tried to implement fetchMore for 250k+ prims (for filtering operations) but
    # it's not as pleasant experience and filtering first on path is a better tactic.
    # https://doc.qt.io/qt-5/qtwidgets-itemviews-fetchmore-example.html
    def __init__(self, columns, *args, **kwargs):
        super().__init__(columns, *args, **kwargs)
        self._stage = None
        self._root_paths = set()
        self._prune_children = set()
        self._filter_predicate = None
        self._traverse_predicate = Usd.PrimAllPrimsPredicate

    @property
    def _prune_predicate(self):
        def prune(prim):
            path = prim.GetPath()
            return any(path.HasPrefix(p) for p in self._prune_children)
        return prune

    @property
    def stage(self):
        return self._stage.__repr__.__self__  # a proxy is currently stored, return the real one when requested

    @stage.setter
    def stage(self, value):
        self.beginResetModel()
        if value:
            prims = _usd.iprims(
                value,
                root_paths=self._root_paths,
                prune_predicate=self._prune_predicate if self._prune_children else None,
                traverse_predicate=self._traverse_predicate
            )
            self._objects = list(filter(self._filter_predicate, prims))
        else:
            self._objects = []
        # TODO: check if this _stage variable can be avoided
        self._stage = weakref.proxy(value) if value and not isinstance(value, weakref.ProxyType) else value
        self.endResetModel()

    def data(self, index:QtCore.QModelIndex, role:int=...) -> typing.Any:
        # Keep consistency with USDView visual style
        if role == QtCore.Qt.ForegroundRole:
            prim = self.data(index, role=_core._QT_OBJECT_DATA_ROLE)
            active = prim.IsActive()
            if prim.IsInstance():
                color = _PrimTextColor.INSTANCE
            elif Usd.Prim.IsInPrototype(prim) or prim.IsInstanceProxy():
                color = _PrimTextColor.PROTOTYPE
            elif (
                prim.HasAuthoredReferences() or
                prim.HasAuthoredPayloads() or
                prim.HasAuthoredInherits() or
                prim.HasAuthoredSpecializes() or
                prim.HasVariantSets()
            ):
                color = _PrimTextColor.ARCS if active else _PrimTextColor.INACTIVE_ARCS
            elif not active:
                color = _PrimTextColor.INACTIVE
            else:
                color = _PrimTextColor.NONE
            return color.value
        elif role == QtCore.Qt.FontRole:
            # Keep similar USDView visual style, just  don't "bold" defined prims.
            prim = self.data(index, role=_core._QT_OBJECT_DATA_ROLE)
            return _prim_font(abstract=prim.IsAbstract(), orphaned=not prim.IsDefined())
        return super().data(index, role)

    def flags(self, index:QtCore.QModelIndex) -> QtCore.Qt.ItemFlags:
        flags = super().flags(index)
        col_index = index.column()
        # only allow edits when:
        if (col_index not in self._locked_columns  # column is unlocked
            and self._columns[col_index].setter  # a setter has been provided
            and not self._objects[index.row()].IsInstanceProxy()  # Not an instance proxy
        ):
            return flags | QtCore.Qt.ItemIsEditable
        return flags


class _Table(_core._ColumnHeaderMixin, QtWidgets.QTableView):
    pass


class _Spreadsheet(QtWidgets.QDialog):
    """TODO:
        - Make paste work with filtered items (paste has been disabled)
        - Setting a prim as instanceable invalidates child prims.
            - Quickest workaround is to invalidate model?
    """
    def __init__(self, model, columns, options: _core._ColumnOptions = _core._ColumnOptions.ALL, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model
        self._columns = columns
        self.table = table = _Table(model, columns, options)
        # TODO: table.setModel(model)

        table.setEditTriggers(QtWidgets.QAbstractItemView.DoubleClicked | QtWidgets.QAbstractItemView.SelectedClicked)
        table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectItems)
        table.setSortingEnabled(True)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(table)
        self.installEventFilter(self)
        self.setLayout(layout)

    def eventFilter(self, source, event):
        if event.type() == QtCore.QEvent.KeyPress and event.matches(QtGui.QKeySequence.Copy):
            self._copySelection()
        elif event.type() == QtCore.QEvent.KeyPress and event.matches(QtGui.QKeySequence.Paste):
            with Sdf.ChangeBlock():
                self._pasteClipboard()
        return super().eventFilter(source, event)

    def _copySelection(self):
        selection = self.table.selectedIndexes()
        if selection:
            rows = sorted(index.row() for index in selection)
            columns = sorted(index.column() for index in selection)
            rowcount = rows[-1] - rows[0] + 1
            colcount = columns[-1] - columns[0] + 1
            table = [[''] * colcount for _ in range(rowcount)]
            for index in selection:
                row = index.row() - rows[0]
                column = index.column() - columns[0]
                table[row][column] = index.data()
            stream = io.StringIO()
            csv.writer(stream, delimiter=csv.excel_tab.delimiter).writerows(table)
            QtWidgets.QApplication.instance().clipboard().setText(stream.getvalue())

    def _pasteClipboard(self):
        logger.warning("Pasting is still experimental")
        text = QtWidgets.QApplication.instance().clipboard().text()
        logger.info(f"Pasting rows with:\n{text}")
        if not text:
            logger.info("Nothing to do!")
            return
        selection_model = self.table.selectionModel()
        selection = selection_model.selectedIndexes()
        logger.info(f"Selection model indexes: {selection}")

        selected_rows = {i.row() for i in selection}
        selected_columns = {i.column() for i in selection}
        # if selection is not continuous, alert the user and abort instead of
        # trying to figure out what to paste where.
        if len(selection) != len(selected_rows) * len(selected_columns):
            msg = ("To paste with multiple selection,"
                   "cells need to be selected continuously\n"
                   "(no gaps between rows / columns).")
            QtWidgets.QMessageBox.warning(self, "Invalid Paste Selection", msg)
            return

        model = self.table.model()
        current_count = model.rowCount()

        logger.info(f"Selected Rows: {selected_rows}")
        logger.info(f"Selected Columns: {selected_columns}")
        logger.info(f"Current count: {current_count}")
        # selection_model.
        selected_row = min(selected_rows, default=current_count)
        selected_column = min(selected_columns, default=0)
        logger.info(f"selected_row, {selected_row}")
        logger.info(f"selected_column, {selected_column}")
        data = tuple(csv.reader(io.StringIO(text), delimiter=csv.excel_tab.delimiter))
        logger.info(f"data, {data}")

        len_data = len(data)
        single_row_source = len_data == 1
        # if data rows are more than 1, we do not allow for gaps on the selected rows to paste.
        if not single_row_source and len_data != len(range(selected_row, max(selected_rows, default=current_count)+1)):
            msg = ("Clipboard data contains multiple rows,"
                   "in order to paste this content, row cells need to be selected continuously\n"
                   "(no gaps between them).")
            QtWidgets.QMessageBox.warning(self, "Invalid Paste Selection", msg)
            return

        orig_sort_enabled = self.table.isSortingEnabled()
        self.table.setSortingEnabled(False)  # prevent auto sort while adding rows

        maxrow = max(selected_row + len(data),  # either the amount of rows to paste
                     max(selected_rows, default=current_count) + 1)  # or the current row count
        logger.debug(f"maxrow, {maxrow}")

        # this is a bit broken. When pasting a single row on N non-sequential selected items,
        # we are pasting the same value on all the inbetween non selected rows. please fix
        def _sourceIndex(index):
            """Recursively get the source index of a proxy model index"""
            source_model = index.model()
            if isinstance(source_model, _core._ProxyModel):
                return _sourceIndex(source_model.mapToSource(index))
            else:
                return index

        for visual_row, rowdata in enumerate(itertools.cycle(data), start=selected_row):
            if visual_row > maxrow:
                logger.debug(f"Visual row {visual_row} maxed maxrow {maxrow}. Stopping. Bye!")
                break
            if single_row_source and visual_row not in selected_rows:
                logger.debug(f"Visual row {visual_row} not in selected rows {selected_rows}. Continue")
                continue
            logger.debug(f"visual_row={visual_row}")
            logger.debug(f"pasting data={rowdata}")
            if visual_row == current_count:
                # TODO: we're at the end of the rows.
                # If we are in a filtered place, alert the user if they want to paste rest
                logger.warning(f">> No more visual rows to paste on: {visual_row}")
            else:
                prim = model.data(model.index(visual_row, 0), _core._QT_OBJECT_DATA_ROLE)
                for column_index, column_data in enumerate(rowdata, start=selected_column):
                    setter = self._columns[column_index].setter
                    if prim:
                        if not setter:
                            logger.debug(f"Skipping since missing setter: {column_data}")
                            continue
                        logger.debug(f"Setting {column_data} with type {type(column_data)} on {prim}")
                        try:
                            setter(prim, column_data)
                        except Exception as exc:
                            logger.debug(exc)
                            import json  # big hack. how to? this happens when pasting boolean types
                            column_data = json.loads(column_data.lower())
                            logger.debug(f"Attempting to parse an {column_data} with type {type(column_data)} on {prim}")
                            setter(prim, column_data)
                    else:
                        s_index = _sourceIndex(model.index(visual_row, column_index))
                        s_item = self.model.itemFromIndex(s_index)
                        s_item.setData(column_data, QtCore.Qt.DisplayRole)

        self.table.setSortingEnabled(orig_sort_enabled)


class _StageSpreadsheet(_Spreadsheet):
    """TODO:
            - Allow to filter via type inheritance
            - Allow setting root paths from selected paths
    """
    def __init__(self, columns, options: _core._ColumnOptions = _core._ColumnOptions.ALL, *args, **kwargs):
        model = StageTableModel(columns=columns)
        super().__init__(model=model, columns=columns, options=options, *args, **kwargs)
        self._model_hierarchy = model_hierarchy = QtWidgets.QPushButton(_core._EMOJI.MODEL_HIERARCHY.value)
        self._filter_predicate = None
        model_hierarchy.setToolTip(
            textwrap.dedent(f"""
                {_core._EMOJI.MODEL_HIERARCHY.value} Valid Model Hierarchy:
    
                The model hierarchy defines a contiguous set of prims descending from a root prim on a stage, all of which are models.
                This means that each prim in the hierarchy is considered "important to consumers".
                """
                    )
        )
        self._instances = instances = QtWidgets.QPushButton(
            _core._EMOJI.INSTANCE_PROXIES.value)
        instances.setToolTip(
            textwrap.dedent(f"""
                {_core._EMOJI.INSTANCE_PROXIES.value} Traverse Instance Proxies:
                
                An instance proxy is a UsdPrim that represents a descendant prim beneath an instance.
                An instance proxy can not be edited. If edits are required, the parent prim makred as "instanceable=True" must
                be updated with "instanceable=False".
                """
                    )
        )
        self._orphaned = orphaned = QtWidgets.QPushButton(_core._EMOJI.ORPHANED.value)
        orphaned.setToolTip(f"{_core._EMOJI.ORPHANED.value} Orphaned Prims")
        self._classes = classes = QtWidgets.QPushButton(_core._EMOJI.CLASSES.value)
        classes.setToolTip(f"{_core._EMOJI.CLASSES.value} Classes (Abstract Prims)")
        self._defined = defined = QtWidgets.QPushButton(_core._EMOJI.DEFINED.value)
        defined.setToolTip(f"{_core._EMOJI.DEFINED.value} Defined Prims")
        self._active = active = QtWidgets.QPushButton(_core._EMOJI.ACTIVE.value)
        active.setToolTip(f"{_core._EMOJI.ACTIVE.value}")
        self._inactive = inactive = QtWidgets.QPushButton(_core._EMOJI.INACTIVE.value)
        inactive.setToolTip(f"{_core._EMOJI.INACTIVE.value}")

        for each in (
        model_hierarchy, instances, orphaned, classes, defined, active, inactive):
            each.setCheckable(True)
            each.clicked.connect(self._update_stage)

        for each in (model_hierarchy, instances, defined, active):
            each.setChecked(True)

        def _buttons_on_frame(*buttons):
            frame = QtWidgets.QFrame()
            layout = QtWidgets.QHBoxLayout()
            layout.setSpacing(0)
            layout.setContentsMargins(0, 0, 0, 0)
            for button in buttons:
                layout.addWidget(button)
            frame.setLayout(layout)
            return frame

        prim_traversal_frame = _buttons_on_frame(model_hierarchy, instances)
        prim_specifier_frame = _buttons_on_frame(orphaned, classes, defined)
        prim_status_frame = _buttons_on_frame(active, inactive)

        hide_key = f"{_core._EMOJI.VISIBILITY.value} Hide All"
        self._vis_states = {f"{_core._EMOJI.VISIBILITY.value} Show All": True, hide_key: False}
        self._vis_key_by_value = {v: k for k, v in self._vis_states.items()}  # True: Show All
        self._vis_all = vis_all = QtWidgets.QPushButton(hide_key)

        lock_key = f"{_core._EMOJI.LOCK.value} Lock All"
        self._lock_states = {lock_key: True, f"{_core._EMOJI.UNLOCK.value} Unlock All": False}
        self._lock_key_by_value = {v: k for k, v in self._lock_states.items()}  # True: Lock All
        self._lock_all = lock_all = QtWidgets.QPushButton(lock_key)

        # table options
        sorting_enabled = QtWidgets.QCheckBox("Sorting")
        sorting_enabled.setChecked(True)

        options_layout = QtWidgets.QHBoxLayout()
        options_layout.addWidget(prim_traversal_frame)
        options_layout.addWidget(prim_specifier_frame)

        self._filters_logical_op = filters_logical_op = QtWidgets.QComboBox()
        for op in (operator.and_, operator.or_):
            op_text = op.__name__.strip("_")
            filters_logical_op.addItem(op_text, userData=op)
        self._filters_logical_op.currentIndexChanged.connect(self._update_stage)

        options_layout.addWidget(filters_logical_op)
        options_layout.addWidget(prim_status_frame)

        options_layout.addWidget(sorting_enabled)
        if _core._ColumnOptions.VISIBILITY in options:
            vis_all.clicked.connect(self._conformVisibility)
            options_layout.addWidget(vis_all)
        if _core._ColumnOptions.LOCK in options:
            lock_all.clicked.connect(self._conformLocked)
            options_layout.addWidget(lock_all)
        options_layout.addStretch()
        sorting_enabled.toggled.connect(self.table.setSortingEnabled)
        self.sorting_enabled = sorting_enabled

        layout = self.layout()
        layout.addLayout(options_layout)
        layout.addWidget(self.table)
        self.setWindowTitle("Spreadsheet Editor")

        # these are here because of the "higher level methods"
        self._column_options = self.table._column_options
        for column_index, column_options in self._column_options.items():
            # TODO: move this to be "public" signals
            column_options._vis_button.clicked.connect(self._conformVisibilitySwitch)
            column_options._lock_button.clicked.connect(self._conformLockSwitch)

    def _conformVisibilitySwitch(self):
        """Make vis option offer inverse depending on how much is currently hidden"""
        counter = Counter(self.table.isColumnHidden(i) for i in self._column_options)
        current, count = next(iter(counter.most_common(1)))
        self._vis_all.setText(self._vis_key_by_value[current])
        return current

    def _conformLockSwitch(self):
        """Make lock option offer inverse depending on how much is currently locked"""
        counter = Counter(widget._lock_button.isChecked() for widget in self._column_options.values())
        current, count = next(iter(counter.most_common(1)))
        self._lock_all.setText(self._lock_key_by_value[not current])
        return current

    def _conformVisibility(self):
        value = self._vis_states[self._vis_all.text()]
        for index, options in self._column_options.items():
            self.table._setColumnVisibility(index, value)
            options._setHidden(not value)

        self._vis_all.setText(self._vis_key_by_value[not value])
        self.table.horizontalHeader()._handleSectionResized(0)

    @_core.wait()
    def _conformLocked(self):
        value = self._lock_states[self._lock_all.text()]
        for index, options in self._column_options.items():
            options._lock_button.setChecked(value)
            self.table._setColumnLocked(index, value)
        self._lock_all.setText(self._lock_key_by_value[not value])

    @property
    def stage(self):
        return self.model.stage

    @stage.setter
    def stage(self, value):
        self._update_stage(stage=value)

    def setStage(self, stage):  # TODO: remove in favor of property
        """Sets the USD stage the spreadsheet is looking at."""
        self._update_stage(stage=stage)

    @_core.wait()
    def _update_stage(self, *args, stage=None):
        filter_predicate = _filter_predicate(
            orphaned=self._orphaned.isChecked(),
            classes=self._classes.isChecked(),
            defined=self._defined.isChecked(),
            active=self._active.isChecked(),
            inactive=self._inactive.isChecked(),
            logical_op=self._filters_logical_op.currentData(QtCore.Qt.UserRole)
        )
        if self._filter_predicate:
            self.model._filter_predicate = lambda prim: self._filter_predicate(prim) and filter_predicate(prim)
        else:
            self.model._filter_predicate = filter_predicate

        self.model._traverse_predicate = _traverse_predicate(
            model_hierarchy=self._model_hierarchy.isChecked(),
            instance_proxies=self._instances.isChecked(),
        )
        self.model.stage = stage if stage else self.model.stage


class SpreadsheetEditor(_StageSpreadsheet):
    def __init__(self, *args, **kwargs):
        def _vis_value(prim):
            # keep homogeneus array by returning a value of the same type for prims
            # that are not imageable (e.g. untyped prims)
            imageable = UsdGeom.Imageable(prim)
            return imageable.GetVisibilityAttr().Get() if imageable else ""

        c = _core._Column
        columns = (
            c("Path", lambda prim: str(prim.GetPath())),
            c("Name", Usd.Prim.GetName),
            c("Asset", lambda prim: Usd.ModelAPI(prim).GetAssetName()),
            c("Type", Usd.Prim.GetTypeName, Usd.Prim.SetTypeName, editor=_prim_type_combobox),
            c("Documentation", Usd.Prim.GetDocumentation, Usd.Prim.SetDocumentation),
            c("Instanceable", Usd.Prim.IsInstance, Usd.Prim.SetInstanceable),
            c("Visibility", _vis_value),
            c("Hidden", Usd.Prim.IsHidden, Usd.Prim.SetHidden),
        )
        super().__init__(columns=columns, *args, **kwargs)



from pathlib import Path
from functools import partial

from pxr import Usd
from grill import cook

from ._qt import QtWidgets, QtCore, QtGui
from . import sheets as _sheets, _graph, _core


class _CreatePrims(QtWidgets.QDialog):
    def __init__(self, columns, *args, **kwargs):
        super().__init__(*args, **kwargs)
        form_l = QtWidgets.QFormLayout()
        layout = QtWidgets.QVBoxLayout()
        form = QtWidgets.QFrame()
        form.setLayout(form_l)
        form_l.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(form)
        self._amount = QtWidgets.QSpinBox()
        self._display_le = QtWidgets.QLineEdit()
        form_l.addRow('📚 Amount:', self._amount)
        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Apply | QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        )
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)

        apply_btn = button_box.button(QtWidgets.QDialogButtonBox.Apply)
        self._applied = apply_btn.clicked

        model = _core.EmptyTableModel(columns=columns)
        self.sheet = sheet = _sheets._Spreadsheet(model, columns, _core._ColumnOptions.NONE)
        self._amount.valueChanged.connect(sheet.model.setRowCount)
        sheet.layout().setContentsMargins(0, 0, 0, 0)

        self._amount.setValue(1)
        self._amount.setMinimum(1)
        self._amount.setMaximum(1000)
        self._splitter = splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        splitter.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding))
        splitter.addWidget(sheet)
        layout.addWidget(splitter)
        layout.addWidget(button_box)
        self.setLayout(layout)
        size = sheet.table.viewportSizeHint()
        size.setWidth(size.width() + 65)  # sensible size at init time
        size.setHeight(self.sizeHint().height())
        self.resize(size)

    @staticmethod
    def _setRepositoryPath(parent=None, caption="Select a repository path"):
        dirpath = QtWidgets.QFileDialog.getExistingDirectory(parent=parent, caption=caption)
        if dirpath:
            token = cook.Repository.set(Path(dirpath))
            print(f"Repository path set to: {dirpath}, token: {token}")
        return dirpath


class CreateAssets(_CreatePrims):
    def __init__(self, *args, **kwargs):
        def _taxon_combobox(parent, option, index):
            combobox = QtWidgets.QComboBox(parent=parent)
            options = sorted(self._taxon_options, key=Usd.Prim.GetName)
            combobox.addItems([p.GetName() for p in options])
            return combobox
        _columns = (
            _core._Column("🧬 Taxon", editor=_taxon_combobox),
            _core._Column("🔖 Name"),
            _core._Column("🏷 Label"),
            _core._Column("📜 Description"),  # TODO: STILL UNUSED
        )

        existing_columns = (_core._Column("🧬 Existing", Usd.Prim.GetName),)
        existing_model = _sheets.StageTableModel(columns=existing_columns)
        existing_model._root_paths = {cook._TAXONOMY_ROOT_PATH}
        existing_model._filter_predicate = lambda prim: prim.GetAssetInfoByKey(cook._ASSETINFO_TAXA_KEY)
        # TODO: turn this into a method to lock all columns?
        existing_model._locked_columns = list(range(len(existing_columns)))
        self._existing_model = existing_model

        super().__init__(_columns, *args, **kwargs)
        self.accepted.connect(self._create)
        self._applied.connect(self._apply)
        self.setWindowTitle("Create Assets")

    @property
    def _taxon_options(self):
        return self._existing_model._objects

    @_core.wait()
    def _create(self):
        if not cook.Repository.get(None):
            if not self._setRepositoryPath(self, "Select a repository path to create assets on"):
                msg = "A repository path must be selected in order to create assets."
                QtWidgets.QMessageBox.warning(self, "Repository path not set", msg)
                return
        # TODO: check for "write._TAXONOMY_ROOT_PATH" existence and handle missing
        root = self._existing_model.stage.GetPrimAtPath(cook._TAXONOMY_ROOT_PATH)
        model = self.sheet.table.model()
        for row in range(model.rowCount()):
            taxon_name = model.data(model.index(row, 0))
            taxon = root.GetPrimAtPath(taxon_name)
            asset_name = model.data(model.index(row, 1))
            if not asset_name:
                # TODO: validate and raise error dialog to user. For now we ignore.
                print(f"An asset name is required! Missing on row: {row}")
                continue
            label = model.data(model.index(row, 2))
            cook.create_unit(taxon, asset_name, label)

    def setStage(self, stage):
        self._existing_model.stage = stage

    def _apply(self):
        """Apply current changes and keep dialog open."""
        # TODO: move this to the base _CreatePrims class
        self._create()
        self.setStage(self._existing_model.stage)


class TaxonomyEditor(_CreatePrims):

    def __init__(self, *args, **kwargs):
        # 1. Read only list of existing taxonomy
        # 2. Place to create new taxon groups
        #    - name | references | id_fields

        class ReferenceSelection(QtWidgets.QDialog):
            def __init__(self, parent=None):
                super().__init__(parent=parent)
                layout = QtWidgets.QVBoxLayout()
                self._options = options = QtWidgets.QListWidget()
                options.setSelectionMode(options.SelectionMode.ExtendedSelection)
                options.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
                options.customContextMenuRequested.connect(lambda: self._create_context_menu().exec_(QtGui.QCursor.pos()))
                layout.addWidget(options)
                button_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
                button_box.accepted.connect(self.accept)
                button_box.rejected.connect(self.reject)
                layout.addWidget(button_box)
                self.setLayout(layout)
                self.setWindowTitle("Extend From...")

            def _create_context_menu(self):
                def set_check_status(status):
                    for each in self._options.selectedItems():
                        each.setCheckState(status)

                menu = QtWidgets.QMenu(self._options)
                for title, status in (
                    ("Check Selected", QtCore.Qt.Checked),
                    ("Uncheck Selected", QtCore.Qt.Unchecked),
                ):
                    menu.addAction(title, partial(set_check_status, status))
                return menu

            def showEvent(self, *args, **kwargs):
                result = super().showEvent(*args, **kwargs)
                # hack? wihout this, we appear off screen (or on top of it)
                self.adjustPosition(self.parent())
                return result

            def property(self, name):  # override ourselves yeahhh
                if name == 'value':
                    return self._value()
                return super().property(name)

            def _value(self):
                return [
                    each.text()
                    for each in self._options.findItems("*", QtCore.Qt.MatchWildcard)
                    if each.checkState() == QtCore.Qt.Checked
                ]

        def _reference_selector(parent, option, index):
            inter = ReferenceSelection(parent=parent)
            inter.setModal(True)
            idata = index.data()
            checked_items = set()
            if idata:
                checked_items.update(idata.split("\n"))

            for taxon in self._taxon_options:
                taxon = taxon.GetName()
                item = QtWidgets.QListWidgetItem(inter._options)
                item.setText(taxon)
                item.setCheckState(QtCore.Qt.Checked if taxon in checked_items else QtCore.Qt.Unchecked)

            return inter

        def _reference_setter(editor: ReferenceSelection, model: _core._ProxyModel, index:QtCore.QModelIndex):
            return model.setData(index, "\n".join(editor._value()))

        identity = lambda x: x
        _columns = (
            _core._Column("🧬 New Name", identity),
            _core._Column(
                "🔗 References",
                identity,
                editor=_reference_selector,
                setter=_reference_setter
            ),
            _core._Column(f"{_core._EMOJI.ID.value} ID Fields", identity),
        )
        super().__init__(_columns, *args, **kwargs)
        self.setWindowTitle("Taxonomy Editor")

        existing_splitter = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        existing_columns = (_core._Column("🧬 Existing", Usd.Prim.GetName),)
        existing_model = _sheets.StageTableModel(columns=existing_columns)
        existing_model._root_paths = {cook._TAXONOMY_ROOT_PATH}
        existing_model._filter_predicate = lambda prim: prim.GetAssetInfoByKey(cook._ASSETINFO_TAXA_KEY)
        # TODO: turn this into a method to lock all columns?
        existing_model._locked_columns = list(range(len(existing_columns)))
        self._existing = existing = _sheets._Spreadsheet(
            existing_model,
            # TODO: see if columns "should" be passed always to the model. If so, then
            #   we can avoid passing it here.
            existing_columns,
            _core._ColumnOptions.SEARCH,
        )
        existing.layout().setContentsMargins(0, 0, 0, 0)
        existing_splitter.addWidget(existing)

        self._graph_view = _graph._GraphViewer(parent=self)
        existing_splitter.addWidget(self._graph_view)

        selectionModel = existing.table.selectionModel()
        selectionModel.selectionChanged.connect(self._existingSelectionChanged)

        self._splitter.insertWidget(0, existing_splitter)
        self._splitter.setStretchFactor(0, 2)
        self._splitter.setStretchFactor(1, 3)

        self.accepted.connect(self._create)
        self._applied.connect(self._apply)

    def _apply(self):
        """Apply current changes and keep dialog open."""
        # TODO: move this to the base _CreatePrims class
        self._create()
        self.setStage(self._existing.model.stage)

    def _existingSelectionChanged(self, selected: QtCore.QItemSelection, deselected: QtCore.QItemSelection):
        prims = (index.data(_core._QT_OBJECT_DATA_ROLE) for index in self._existing.table.selectedIndexes())
        self._graph_view.view([prim.GetName() for prim in prims])

    @property
    def _taxon_options(self):
        return self._existing.model._objects

    def setStage(self, stage):
        self._existing.model.stage = stage
        existing_taxa = self._taxon_options
        self._graph_view.graph = cook.taxonomy_graph(existing_taxa, self._graph_view.url_id_prefix)

    @_core.wait()
    def _create(self):
        if not cook.Repository.get(None):
            if not self._setRepositoryPath(self, "Select a repository path to create assets on"):
                msg = "A repository path must be selected in order to create assets."
                QtWidgets.QMessageBox.warning(self, "Repository path not set", msg)
                return
        # TODO: check for "write._TAXONOMY_ROOT_PATH" existence and handle missing
        # TODO: make data point to the actual existing prims from the start.
        #   So that we don't need to call GetPRimAtPath.
        model = self.sheet.table.model()
        stage = self._existing.model.stage
        root = stage.GetPrimAtPath(cook._TAXONOMY_ROOT_PATH)
        for row in range(model.rowCount()):
            taxon_name = model.data(model.index(row, 0))
            if not taxon_name:
                # TODO: validate and raise error dialog to user. For now we ignore.
                print(f"An asset name is required! Missing on row: {row}")
                continue
            reference_names = (model.data(model.index(row, 1)) or '').split("\n")
            references = (root.GetPrimAtPath(ref_name) for ref_name in reference_names if ref_name)
            cook.define_taxon(stage, taxon_name, references=references)



from maya import cmds
from PySide2 import QtCore


def install():
    # After trial and error, it looks like waiting for a bit via single shot
    # guarantees that the deferred command will execute, even with 1 millisecond.
    QtCore.QTimer.singleShot(1, lambda: cmds.evalDeferred(_install, lp=True))


def _install():
    if cmds.about(batch=True):
        return  # No GUI, nothing to install
    from ..views import maya
    maya.create_menu()



"""This module aims to provide ease startup of The Grill menus and plugins."""
import os
from pathlib import Path


def install(sitedir):
    # NOTE: if plugInfo provides more plugins other than USDView, it will need to be
    # executed in deferred evaluation in maya and extended on the houdini package.
    _usd_pluginfo(sitedir)

    if "MAYA_PLUG_IN_PATH" in os.environ:  # Quick check for maya plugins
        try:
            from . import maya
        except ImportError:  # Not a proper Maya environment
            pass
        else:
            maya.install()


def _usd_pluginfo(sitedir):
    os.environ["PXR_PLUGINPATH_NAME"] = f"{Path(sitedir) / 'grill' / 'resources' / 'usd' / 'plugInfo.json'}{os.pathsep}{os.environ.get('PXR_PLUGINPATH_NAME', '')}"



import hou
import toolutils
from functools import cache, lru_cache, partial
from . import sheets as _sheets, description as _description, create as _create, stats as _stats, _core
_description._PALETTE.set(0)  # (0 == dark, 1 == light)
_core._ensure_dot()


def _stage_on_widget(widget_creator, _cache=True):
    _description._PALETTE.set(0)  # (0 == dark, 1 == light)
    def _launcher():
        widget = widget_creator(parent=hou.qt.mainWindow())
        # TODO: Get stage from current node or from current viewport?
        #       Investigate pros & cons and decide.
        viewer = toolutils.sceneViewer()
        stage = viewer.stage()
        if stage:
            widget.setStage(stage)

        return widget

    if _cache:
        _launcher = cache(_launcher)
    return _launcher


@lru_cache(maxsize=None)
def _creator(widget_creator):
    widget = _stage_on_widget(widget_creator)()
    def refresh_ui():
        viewer = toolutils.sceneViewer()
        node = viewer.currentNode()
        node.cook(force=True)

    widget.accepted.connect(refresh_ui)
    # widget._applied.connect(refresh_ui)  # TODO: fix, errors with PySide Qt 5.12 (houdini)
    return widget


@cache
def _spreadsheet():
    """This is meant to be run under a solaris desktop in houdini.

    :return:
    """
    widget = _stage_on_widget(_sheets.SpreadsheetEditor)()
    def refresh_ui():
        viewer = toolutils.sceneViewer()
        node = viewer.currentNode()
        node.cook(force=True)

    # https://github.com/nexdatas/selector/blob/b4d6ae92f8887d1bee8bd45dd3cc5777be5ccceb/nxsselector/ElementModel.py#L501
    # https://github.com/pyblish/pyblish-qml/blob/master/pyblish_qml/models.py
    # widget.table.itemChanged.connect(refresh_ui)  # TODO: disabled since new model does not have refresh capabilities yet!
    return widget


def _prim_composition():
    _description._PALETTE.set(0)  # (0 == dark, 1 == light)
    editor = _description.PrimComposition(parent=hou.qt.mainWindow())
    editor._prim = None

    def _updatePrim():
        # find a cheaper way for this?
        viewer = toolutils.sceneViewer()
        stage = viewer.stage()
        if not stage:
            editor.clear()
            editor._prim = None
            return
        selection = viewer.currentSceneGraphSelection()
        prims = tuple(stage.GetPrimAtPath(path) for path in selection)
        prim = next(iter(prims), None)
        if not prim:
            if editor._prim:
                editor.clear()
                editor._prim = None
        else:
            if prim != editor._prim:
                editor.setPrim(prim)
                editor._prim = prim

    hou.ui.addEventLoopCallback(_updatePrim)
    return editor


def _connectable_viewer():
    _description._PALETTE.set(0)  # (0 == dark, 1 == light)
    editor = _description._ConnectableAPIViewer(parent=hou.qt.mainWindow())
    editor._prim = None

    def _updatePrim():
        # find a cheaper way for this?
        viewer = toolutils.sceneViewer()
        stage = viewer.stage()
        if not stage:
            editor.setPrim(None)
            return
        selection = viewer.currentSceneGraphSelection()
        prims = tuple(stage.GetPrimAtPath(path) for path in selection)
        prim = next(iter(prims), None)
        if not prim:
            if editor._prim:
                editor.setPrim(None)
                editor._prim = None
        else:
            if prim != editor._prim:
                editor.setPrim(prim)
                editor._prim = prim

    hou.ui.addEventLoopCallback(_updatePrim)
    return editor


_create_assets = partial(_creator, _create.CreateAssets)
_taxonomy_editor = partial(_creator, _create.TaxonomyEditor)
_layerstack_composition = _stage_on_widget(_description.LayerStackComposition)
_stage_stats = _stage_on_widget(_stats.StageStats, _cache=False)



try:  # only while transition from PySide2 to PySide6 happens
    from PySide6 import QtWidgets, QtGui, QtCore, QtCharts, QtSvg, QtTest
except ImportError:
    from PySide2 import QtWidgets, QtGui, QtCore, QtSvg, QtTest
    if not hasattr(QtCore, "__enter__"):
        class SignalBlocker(QtCore.QSignalBlocker):
            def __enter__(self):
                yield

            def __exit__(self, exc_type, exc_value, exc_traceback):
                self.unblock()

        QtCore.QSignalBlocker = SignalBlocker
    try:
        from PySide2.QtCharts import QtCharts
    except ImportError:
        # Maya-2023.3 and Houdini-19.5.534 bundle PySide2, but their versions fail to bring QtCharts :c
        pass



import unittest
from grill.logger import model


class TestLogger(unittest.TestCase):

    def test_root_logger(self):
        from grill.logger import LOG
        LOG.info("Info message")

    def test_module_logger(self):
        from grill.logger import get_logger
        logger = get_logger(__name__)
        logger.info("Testing inner module")

    def test_default(self):
        l = model.LogFile.get_default(year=1999, day=1)
        self.assertEqual(l.year, '1999')

        with self.assertRaises(ValueError):
            l.month = 13  # invalid month 13

        for m in range(1, 13):
            l.month = m
